#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# import crcmod
import google_crc32c
import requests
import re
import os
import time
import base64
import ast
import platform
import importlib.metadata
import logging
import warnings
import pandas as pd
from tqdm import tqdm
from .metadata_helpers import get_recent

logging.basicConfig(level=logging.INFO, format="%(message)s")

# Set global user agent
vers = importlib.metadata.version("neonutilities")
plat = platform.python_version()
osplat = platform.platform()

usera = f"neonutilities/{vers} Python/{plat} {osplat}"


def token_date(token, rval="string"):
    """

    Find the expiration date of a user-specific API token generated within data.neonscience.org user accounts.

    Parameters
    --------
    token: User specific API token (generated within data.neonscience.org user accounts).
    rval: Should returned value be a string or a time value? Defaults to string.

    Return
    --------
    Date of expiration in local time
    
    Created on Jan 13 2026

    @author: Claire Lunch
    """
    
    splittoken = token.split(".")
    # + "===" added here as "padding" required to make all tokens interpretable
    dubsplit = base64.urlsafe_b64decode(splittoken[1] + "===")
    dictsplit = ast.literal_eval(dubsplit.decode("utf-8"))
    if "exp" in dictsplit.keys():
        expsplit = dictsplit["exp"]
        if rval=="string":
            expdate = time.asctime(time.localtime(expsplit))
            return(expdate)
        else:
            return(expsplit)
    else:
        logging.info("No expiration date found for API token.")
        return None


def token_check(token):
    """

    Check whether a NEON API token has expired.

    Parameters
    --------
    token: User specific API token (generated within data.neonscience.org user accounts).

    Return
    --------
    The original token, if unexpired. If expired, return None.
    
    Created on Jan 13 2026

    @author: Claire Lunch
    """
    
    try:
        expdate = token_date(token, rval="time")
    except Exception:
        logging.info("API token expiration date could not be determined.")
        return(token)
    
    if(expdate is None):
        return(token)
    else:
        if time.time() > expdate:
            logging.info("API token has expired. Function will proceed using public access rate. Go to your NEON user account to generate a new token.")
            token = None
        return(token)
    

def get_api(api_url, token=None):
    """

    Accesses the API with options to use the user-specific API token generated within data.neonscience.org user accounts.

    Parameters
    --------
    api_url: The API endpoint URL.
    token: User specific API token (generated within data.neonscience.org user accounts). Optional.

    Return
    --------
    API GET response containing status code and data that can be parsed into a json file

    Example
    --------
    Get the sample class for a surface water particulate stable isotope sample

    >>> sample_get = get_api(api_url='https://data.neonscience.org/api/v0/samples/classes?sampleTag=MCRA.SS.20230425.POM.1',token=None)

    Created on Fri Aug 30 2023

    @author: Zachary Nickerson
    """

    def get_status_code_meaning(status_code):
        return requests.status_codes._codes[status_code][0]

    # Check internet connection
    try:
        check_connection = requests.get(
            "https://data.neonscience.org/", headers={"User-Agent": usera}
        )
        if check_connection.status_code != 200:
            status_code = check_connection.status_code
            status_code_meaning = get_status_code_meaning(status_code)
            raise ConnectionError(
                f"Request failed with status code {status_code}, indicating '{status_code_meaning}'\n"
            )
    except Exception:  # ConnectionError as e
        raise ConnectionError("Connection error. Cannot access NEON API.\n")

    # Make 5 request attempts. If the rate limit is reached, pause for the
    # burst reset time to try again.
    j = 1

    while j <= 5:
        # Try making the request
        try:
            # Construct URL either with or without token
            if token is None:
                response = requests.get(
                    api_url, headers={"accept": "application/json", "User-Agent": usera}
                )
            else:
                response = requests.get(
                    api_url,
                    headers={
                        "X-API-TOKEN": token,
                        "accept": "application/json",
                        "User-Agent": usera,
                    },
                )

            # Check for successful response
            if response.status_code == 200:
                if "x-ratelimit-limit" in response.headers:
                    # Retry get request if rate limit is reached
                    limit_remain = response.headers.get("x-ratelimit-remaining")

                    if int(limit_remain) < 1:
                        # Wait for the reset time
                        time_reset = response.headers.get("x-ratelimit-reset")
                        logging.info(
                            f"Rate limit reached. Pausing for {time_reset} seconds to reset.\n"
                        )
                        time.sleep(int(time_reset))
                        # Increment loop to retry request attempt
                        j += 1

                    else:
                        # If rate limit is not reached, exit out of loop
                        j += 5

                else:
                    # If x-ratelimit-limit not found in headers, exit out of
                    # loop (don't need to retry because of rate-limit)
                    j += 5

            else:
                # Return nothing if request failed (status code is not 200)
                # Print the status code and it's meaning
                status_code_meaning = get_status_code_meaning(response.status_code)
                raise ConnectionError(
                    f"Request failed with status code {response.status_code}, indicating '{status_code_meaning}'\n"
                )

            return response

        except Exception as error:
            print(error)
            return None


def get_api_headers(api_url, token=None):
    """

    Accesses the API with options to use the user-specific API token generated within neon.datascience user accounts.

    Parameters
    --------
    api_url: The API endpoint URL.
    token: User specific API token (generated within neon.datascience user accounts). Optional.

    Return
    --------
    The header only from an API GET response

    Created on Feb 26 2024

    Adapted from get_api
    @author: Zachary Nickerson
    @author: Claire Lunch
    """

    def get_status_code_meaning(status_code):
        return requests.status_codes._codes[status_code][0]

    # Check internet connection
    try:
        check_connection = requests.head(
            "https://data.neonscience.org/", headers={"User-Agent": usera}
        )
        if check_connection.status_code != 200:
            status_code = check_connection.status_code
            status_code_meaning = get_status_code_meaning(status_code)
            raise ConnectionError(
                f"Request failed with status code {status_code}, indicating '{status_code_meaning}'\n"
            )
    except Exception:  # ConnectionError as e
        raise ConnectionError(
            "No internet connection detected. Cannot access NEON API.\n"
        )

    # Make 5 request attempts. If the rate limit is reached, pause for the
    # burst reset time to try again.
    j = 1

    while j <= 5:
        # Try making the request
        try:
            # Construct URL either with or without token
            if token is None:
                response = requests.head(
                    api_url, headers={"accept": "application/json", "User-Agent": usera}
                )
            else:
                response = requests.head(
                    api_url,
                    headers={
                        "X-API-TOKEN": token,
                        "accept": "application/json",
                        "User-Agent": usera,
                    },
                )

            # Check for successful response
            if response.status_code == 200:
                if "x-ratelimit-limit" in response.headers:
                    # Retry get request if rate limit is reached
                    limit_remain = response.headers.get("x-ratelimit-remaining")

                    if int(limit_remain) < 1:
                        # Wait for the reset time
                        time_reset = response.headers.get("x-ratelimit-reset")
                        logging.info(
                            f"Rate limit reached. Pausing for {time_reset} seconds to reset.\n"
                        )
                        time.sleep(int(time_reset))
                        # Increment loop to retry request attempt
                        j += 1

                    else:
                        # If rate limit is not reached, exit out of loop
                        j += 5

                else:
                    # If x-ratelimit-limit not found in headers, exit out of
                    # loop (don't need to retry because of rate-limit)
                    j += 5

            else:
                # Return nothing if request failed (status code is not 200)
                # Print the status code and it's meaning
                status_code_meaning = get_status_code_meaning(response.status_code)
                raise ConnectionError(
                    f"Request failed with status code {response.status_code}, indicating '{status_code_meaning}'\n"
                )

            return response

        except Exception:
            raise ConnectionError(
                "No response. NEON API may be unavailable, check NEON data portal for outage alerts. If the problem persists and can't be traced to an outage alert, check your computer for firewall or other security settings preventing Python from accessing the internet."
            )


def get_zip_urls(
    url_set, package, release, include_provisional, token=None, progress=True
):
    """

    Given a set of urls to the data endpoint of the NEON API, returns the set of zip file urls for each site-month package. Internal function, called by zips_by_product().

    Parameters
    --------
    url_set: A list of urls pointing to the data endpoint of the NEON API
    package: Data download package, basic or expanded.
    release: Data release to download.
    include_provisional: Should Provisional data be returned in the download?
    token: User specific API token (generated within neon.datascience user accounts). Optional.
    progress: Should the progress bar be displayed?

    Return
    --------
    List of urls pointing to zip files for each product-site-month.

    Created on Feb 23 2024

    @author: Claire Lunch
    """

    flnm = []
    z = []
    sz = []
    rel = []
    provflag = False
    if progress:
        logging.info("Finding available files")

    for i in tqdm(range(0, len(url_set)), disable=not progress):
        # get list of files from data endpoint
        m_res = get_api(api_url=url_set[i], token=token)
        if m_res is None:
            logging.info(
                "Connection error for a subset of urls. Check outputs for missing data."
            )
            return None
        m_di = m_res.json()

        # only keep queried release
        if release != "current":
            if release != m_di["data"]["release"]:
                continue

        # if include_provisional=F, exclude provisional
        if not include_provisional:
            if m_di["data"]["release"] == "PROVISIONAL":
                provflag = True
                continue

        # check for no files
        if "packages" not in list(m_di["data"]):
            logging.info(
                f"No files found for site {m_di['data']['siteCode']} and month {m_di['data']['month']}"
            )
            continue

        if len(m_di["data"]["packages"]) == 0:
            logging.info(
                f"No files found for site {m_di['data']['siteCode']} and month {m_di['data']['month']}"
            )
            continue

        # if package=expanded, check for expanded. reassign to basic if not found.
        if package == "expanded":
            if package not in [p["type"] for p in m_di["data"]["packages"]]:
                logging.info(
                    f"No expanded package found for site {m_di['data']['siteCode']} and month {m_di['data']['month']}. Basic package downloaded instead."
                )
                package = "basic"

        # get zip file url and file name
        zi = [u["url"] for u in m_di["data"]["packages"] if u["type"] == package]
        h = get_api_headers(api_url=zi[0], token=token)
        fltp = re.sub(pattern='"', repl="", string=h.headers["content-disposition"])
        flnmi = re.sub(pattern="inline; filename=", repl="", string=fltp)

        # get file sizes
        szr = re.compile(package)
        flszs = [siz["size"] for siz in m_di["data"]["files"] if szr.search(siz["url"])]
        flszi = sum(flszs)

        # return url, file name, file size, and release
        flnm.append(flnmi)
        z.append(zi)
        sz.append(flszi)
        rel.append(m_di["data"]["release"])

    z = sum(z, [])
    zpfiles = dict(flnm=flnm, z=z, sz=sz, rel=rel)

    # provisional message
    if provflag:
        logging.info(
            "Provisional data were excluded from available files list. To download provisional data, use input parameter include_provisional=True."
        )

    return zpfiles


def get_tab_urls(
    url_set,
    package,
    release,
    include_provisional,
    timeindex,
    tabl,
    token=None,
    progress=True,
):
    """

    Given a set of urls to the data endpoint of the NEON API, and averaging interval or table name criteria, returns the set of urls to individual files for each site-month package. Internal function, called by zips_by_product().

    Parameters
    --------
    url_set: A list of urls pointing to the data endpoint of the NEON API
    package: Data download package, basic or expanded.
    release: Data release to download.
    include_provisional: Should Provisional data be returned in the download?
    timeindex: Averaging interval of data to download.
    tabl: Table name of data to download.
    token: User specific API token (generated within neon.datascience user accounts). Optional.
    progress: Should the progress bar be displayed?

    Return
    --------
    List of urls pointing to files for each product-site-month and subset.

    Created on Mar 23 2024

    @author: Claire Lunch
    """

    # initiate file lists
    flnm = []
    flpth = []
    z = []
    sz = []
    rel = []
    varf = []
    rdme = []
    sp = []

    # create regular expressions for file finding
    vr = re.compile("variables")
    rdr = re.compile("readme")
    spr = re.compile("sensor_positions")

    if timeindex != "all":
        tt = re.compile(
            str(timeindex) + "min|" + str(timeindex) + "_min|science_review_flags"
        )

    if tabl != "all":
        tb = re.compile("[.]" + tabl + "[.]")

    provflag = False
    if progress:
        logging.info("Finding available files")

    for i in tqdm(range(0, len(url_set)), disable=not progress):
        # get list of files from data endpoint
        m_res = get_api(api_url=url_set[i], token=token)
        if m_res is None:
            logging.info(
                "Connection error for a subset of urls. Check outputs for missing data."
            )
            return None
        m_di = m_res.json()

        # only keep queried release
        if release != "current":
            if release != m_di["data"]["release"]:
                continue

        # if include_provisional=F, exclude provisional
        if not include_provisional:
            if m_di["data"]["release"] == "PROVISIONAL":
                provflag = True
                continue

        # subset to package. switch to basic if expanded not available
        # package name isn't always in file name (lab files, SRFs) but is always in url
        pr = re.compile(package)
        flsp = [f for f in m_di["data"]["files"] if pr.search(f["url"])]
        if package == "expanded" and len(flsp) == 0:
            pr = re.compile("basic")
            flsp = [f for f in m_di["data"]["files"] if pr.search(f["url"])]

        # check for no files
        if len(flsp) == 0:
            logging.info(
                f"No files found for site {m_di['data']['siteCode']} and month {m_di['data']['month']}"
            )
            continue

        # get zip file url and file name
        zi = [u["url"] for u in m_di["data"]["packages"] if u["type"] == package]
        h = get_api_headers(api_url=zi[0], token=token)
        fltp = re.sub(pattern='"', repl="", string=h.headers["content-disposition"])
        flpthit = re.sub(pattern="inline; filename=", repl="", string=fltp)
        flpthi = re.sub(pattern=".zip", repl="/", string=flpthit)

        # make separate lists of variables, readme and sensor positions
        varfi = [f for f in m_di["data"]["files"] if vr.search(f["name"])]
        rdmei = [f for f in m_di["data"]["files"] if rdr.search(f["name"])]
        spi = [f for f in m_di["data"]["files"] if spr.search(f["name"])]
        for f in varfi:
            f["name"] = flpthi + f["name"]
        for f in rdmei:
            f["name"] = flpthi + f["name"]

        varf.append(varfi)
        rdme.append(rdmei)
        if len(spi) > 0:
            for f in spi:
                f["name"] = flpthi + f["name"]
            sp.append(spi)

        # subset by averaging interval
        if timeindex != "all":
            flnmi = [flpthi + fl["name"] for fl in flsp if tt.search(fl["name"])]
            flszi = [fl["size"] for fl in flsp if tt.search(fl["name"])]
            zi = [fl["url"] for fl in flsp if tt.search(fl["name"])]

            # check for no files
            if len(flnmi) == 0:
                logging.info(
                    f"No files found for site {m_di['data']['siteCode']}, month {m_di['data']['month']}, and averaging interval (time index) {timeindex}"
                )
                continue

        # subset by table
        if tabl != "all":
            flnmi = [flpthi + fl["name"] for fl in flsp if tb.search(fl["name"])]
            flszi = [fl["size"] for fl in flsp if tb.search(fl["name"])]
            zi = [fl["url"] for fl in flsp if tb.search(fl["name"])]

            # check for no files
            if len(flnmi) == 0:
                logging.info(
                    f"No files found for site {m_di['data']['siteCode']}, month {m_di['data']['month']}, and table {tabl}"
                )
                continue

        # return url, file name, file size, and release
        flnm.append(flnmi)
        flpth.append(flpthi)
        z.append(zi)
        sz.append(flszi)
        rel.append(m_di["data"]["release"])

    # get most recent metadata files from lists
    try:
        varf = sum(varf, [])
        varfl = get_recent(varf, "variables")
        flnm.append([fl["name"] for fl in varfl])
        z.append([fl["url"] for fl in varfl])
        sz.append([fl["size"] for fl in varfl])
    except Exception:
        pass

    try:
        rdme = sum(rdme, [])
        rdfl = get_recent(rdme, "readme")
        flnm.append([fl["name"] for fl in rdfl])
        z.append([fl["url"] for fl in rdfl])
        sz.append([fl["size"] for fl in rdfl])
    except Exception:
        pass

    # get most recent sensor positions file for each site
    if len(sp) > 0:
        sp = sum(sp, [])
        sr = re.compile("[/][A-Z]{4}[/]")
        sites = [sr.search(f["url"]).group(0) for f in sp]
        sites = list(set(sites))
        sites = [re.sub(pattern="/", repl="", string=s) for s in sites]
        try:
            for s in sites:
                spfl = get_recent(sp, s)
                flnm.append([fl["name"] for fl in spfl])
                z.append([fl["url"] for fl in spfl])
                sz.append([fl["size"] for fl in spfl])
        except Exception:
            pass

    z = sum(z, [])
    flnm = sum(flnm, [])
    sz = sum(sz, [])
    tbfiles = dict(flnm=flnm, flpth=flpth, z=z, sz=sz, rel=rel)

    # provisional message
    if provflag:
        logging.info(
            "Provisional data were excluded from available files list. To download provisional data, use input parameter include_provisional=True."
        )

    return tbfiles


def download_urls(url_set, outpath, token=None, progress=True):
    """

    Given a set of urls to NEON data packages or files, downloads the contents of each. Internal function, called by zips_by_product().

    Parameters
    --------
    url_set: A list of urls pointing to zipped data packages
    outpath: Filepath of the folder to download to
    token: User specific API token (generated within neon.datascience user accounts). Optional.
    progress: Should the progress bar be displayed?

    Return
    --------
    Files in the designated folder

    Created on Feb 28 2024

    @author: Claire Lunch
    """

    if progress:
        logging.info("Downloading files")

    for i in tqdm(range(0, len(url_set["z"])), disable=not progress):
        if len(outpath + url_set["flnm"][i]) > 260 and platform.system() == "Windows":
            raise OSError(
                f'Filepath is {len(outpath+url_set["flnm"][i])} characters long. Filepaths on Windows are limited to 260 characters. Move your working directory closer to the root directory or enable long path support in Windows through the Registry Editor.'
            )
            return

        else:
            try:
                if token is None:
                    j = 0
                    while j < 3:
                        try:
                            with open(outpath + url_set["flnm"][i], "wb") as out_file:
                                content = requests.get(
                                    url_set["z"][i],
                                    stream=True,
                                    headers={
                                        "accept": "application/json",
                                        "User-Agent": usera,
                                    },
                                    timeout=(10, 120),
                                ).content
                                out_file.write(content)
                            j = j + 5
                        except Exception as e:
                            logging.info(
                                f"File {url_set['flnm'][i]} could not be downloaded. Re-attempting."
                            )
                            print(e)
                            j = j + 1
                            time.sleep(5)
                else:
                    j = 0
                    while j < 3:
                        try:
                            with open(outpath + url_set["flnm"][i], "wb") as out_file:
                                content = requests.get(
                                    url_set["z"][i],
                                    stream=True,
                                    headers={
                                        "X-API-TOKEN": token,
                                        "accept": "application/json",
                                        "User-Agent": usera,
                                    },
                                    timeout=(10, 120),
                                ).content
                                out_file.write(content)
                            j = j + 5
                        except Exception as e:
                            logging.info(
                                f"File {url_set['flnm'][i]} could not be downloaded. Re-attempting."
                            )
                            print(e)
                            j = j + 1
                            time.sleep(5)

            except Exception:
                logging.info(
                    f"File {url_set['flnm'][i]} could not be downloaded and was skipped. If this issue persists, check your network connection and check the NEON Data Portal for outage alerts."
                )
                pass

    return None


# def calculate_crc32c(filename):
#     """
#     This function calculates the CRC32C checksum of a file. This is designed to use
#     in the check_exists_and_checksum function to .

#     Parameters
#     ----------
#     filename : str
#         Path to the file for which to calculate the CRC32C checksum.

#     Returns
#     -------
#     str
#         The CRC32C checksum as an 8-character hexadecimal string (zero-padded if necessary).

#     Example
#     -------
#     >>> calculate_crc32c("myfile.tif")
#     '036d153e'
#     """
#     crc32c_func = crcmod.predefined.mkPredefinedCrcFun('crc-32c')
#     with open(filename, 'rb') as f:
#         return format(crc32c_func(f.read()), '08x')


def calculate_crc32c(filename, chunk_size=1024 * 1024):
    """
    Faster CRC32C checksum using google-crc32c.
    """
    crc32c = google_crc32c.Checksum()
    with open(filename, "rb") as f:
        while True:
            chunk = f.read(chunk_size)
            if not chunk:
                break
            crc32c.update(chunk)
    return crc32c.digest().hex().zfill(8)


def download_file(url, savepath, chunk_size=1024, token=None):
    """
    This function downloads a single file from a Google Cloud Storage URL to a user-specified directory.

    Parameters
    --------
    url: str
        The Google Cloud Storage URL where the file is stored.

    savepath: str
        The file location (path) where the file will be downloaded.

    chunk_size:
        Size in bytes of chunks for chunked download

    token: str, optional
        User-specific API token generated within neon.datascience user accounts. If provided, it will be used for authentication.

    Returns
    --------
    None

    Raises
    --------
    None

    Examples
    --------
    >>> download_file('https://storage.googleapis.com/neon-aop-products/2023/FullSite/D02/2023_SCBI_6/L3/DiscreteLidar/CanopyHeightModelGtif/NEON_D02_SCBI_DP3_741000_4301000_CHM.tif', '/path/to/save', 'my-api-token')
    # This will download 'NEON_D02_SCBI_DP3_741000_4301000_CHM.tif' from the specified URL to '/path/to/save' directory using 'my-api-token' for authentication.

    Notes
    --------
    The function creates the directory specified by 'savepath' if it does not exist.
    It also downloads the readme.txt file which contains detailed information about the data package, issue logs, etc.
    https://storage.googleapis.com/neon-publication/NEON.DOM.SITE.DP3.30015.001/SCBI/20230601T000000--20230701T000000/basic/NEON.D02.SCBI.DP3.30015.001.readme.20240206T001418Z.txt

    The function issues a warning on Windows systems if the full download file path exceeds 260 characters, as the file may not be downloaded due to path length limitations.

    @author: Bridget Hass

    """

    pathparts = url.split("/")
    file_path = "/".join(pathparts[3 : len(pathparts)])

    file_fullpath = savepath + "/" + file_path
    file_fullpath_abs = os.path.abspath(file_fullpath)  # get the absolute path

    if len(file_fullpath_abs) > 260 and platform.system() == "Windows":
        warnings.warn(
            f"Filepaths on Windows are limited to 260 characters. Attempting to download a filepath that is {len(file_fullpath_abs)} characters long. Set the working or savepath directory to be closer to the root directory or enable long path support in Windows."
        )
        # return

    else:
        os.makedirs(os.path.dirname(file_fullpath), exist_ok=True)

        try:
            if token is None:
                j = 0
                while j < 3:
                    try:
                        r = requests.get(
                            url,
                            stream=True,
                            headers={"accept": "application/json", "User-Agent": usera},
                            timeout=(10, 120),
                        )
                        j = j + 5
                    except Exception:
                        logging.info(
                            f"File {os.path.basename(url)} could not be downloaded. Re-attempting."
                        )
                        j = j + 1
                        time.sleep(5)
            else:
                j = 0
                while j < 3:
                    try:
                        r = requests.get(
                            url,
                            stream=True,
                            headers={
                                "X-API-TOKEN": token,
                                "accept": "application/json",
                                "User-Agent": usera,
                            },
                            timeout=(10, 120),
                        )
                        j = j + 5
                    except Exception:
                        logging.info(
                            f"File {os.path.basename(url)} could not be downloaded. Re-attempting."
                        )
                        j = j + 1
                        time.sleep(5)

            with open(file_fullpath, "wb") as f:
                for chunk in r.iter_content(chunk_size=chunk_size):
                    if chunk:
                        f.write(chunk)
            r.close()

        except Exception:
            logging.info(
                f"File {os.path.basename(url)} could not be downloaded and was skipped or partially downloaded. If this issue persists, check your network connection and check the NEON Data Portal for outage alerts."
            )
            # print(e)
            pass

        return


def readme_url(readmepath):
    """
    This function accesses a readme file via a url and converts to a usable format.
    Called by the stacking functions when run in cloud mode.
    A token is not used in the API call since tokens are not used in the stacking functions.

    Parameters
    --------
    readmepath: str
        URL path to the readme file.

    Returns
    --------
    DataFrame
        A pandas data frame of the readme content.
    """

    rdres = requests.get(
        readmepath, headers={"accept": "application/json", "User-Agent": usera}
    )
    rdtxt = rdres.text
    rdlst = rdtxt.split("\n")
    rdfrm = pd.DataFrame(rdlst)

    return rdfrm



# -*- coding: utf-8 -*-
"""
Created on 14 Mar 2024

@author: Claire Lunch

Unit tests for get_citation()

Mocking is not used here, tests access API resources.

"""

# import required packages
import os
from datetime import date
from src.neonutilities.citation import get_citation

# read in token from os.environ
token = os.environ.get("NEON_TOKEN")

# Test a provisional and a release citation


def test_get_citation_provisional():
    """
    Test that the get_citation() function returns the expected citation for provisional data
    """
    cit = get_citation(dpid="DP1.10003.001", release="PROVISIONAL")
    dt = date.today()
    year = dt.year
    citexp = "@misc{DP1.10003.001/provisional,\n  doi = {},\n  url = {https://data.neonscience.org/data-products/DP1.10003.001},\n  author = {{National Ecological Observatory Network (NEON)}},\n  language = {en},\n  title = {Breeding landbird point counts (DP1.10003.001)},\n  publisher = {National Ecological Observatory Network (NEON)},\n  year = {" + str(year) + "}\n}"
    assert cit == citexp


def test_get_citation_release():
    """
    Test that the get_citation() function returns the expected citation for a Release
    """
    cit = get_citation(dpid="DP1.10098.001", release="RELEASE-2023")
    citexp = "@misc{https://doi.org/10.48443/73zn-k414,\n  doi = {10.48443/73ZN-K414},\n  url = {https://data.neonscience.org/data-products/DP1.10098.001/RELEASE-2023},\n  author = {{National Ecological Observatory Network (NEON)}},\n  keywords = {plant productivity, production, carbon cycle, biomass, vegetation, productivity, plants, trees, shrubs, lianas, saplings, net primary productivity (NPP), annual net primary productivity (ANPP), woody plants, vegetation structure, tree height, canopy height, vst},\n  language = {en},\n  title = {Vegetation structure (DP1.10098.001)},\n  publisher = {National Ecological Observatory Network (NEON)},\n  year = {2023}\n}\n"
    assert cit == citexp



# -*- coding: utf-8 -*-
"""
Created on Mon Apr  8 10:32:58 2024

@author: bhass

Get the issue log for a specific data product.

Adapted from R neonUtilities getIssueLog.R
https://github.com/NEONScience/NEON-utilities/blob/main/neonUtilities/R/getIssueLog.R
written by:
@author: Claire Lunch (clunch@battelleecology.org)

"""

import re
import logging
import pandas as pd
from .helper_mods.api_helpers import get_api

logging.basicConfig(level=logging.INFO, format="%(message)s")

# %% functions to validate inputs (should pull these out into another helper module??)


def validate_dpid(dpid):
    """
    Validates the format of a NEON data product ID (dpid).

    Args:
        dpid (str): The NEON data product ID to validate.

    Raises:
        ValueError: If the DPID is not in the correct format.
    """
    dpid_pattern = "DP[1-4]{1}.[0-9]{5}.00[1-2]{1}"
    if not re.fullmatch(dpid_pattern, dpid):
        raise ValueError(
            f"{dpid} is not a properly formatted data product ID. The correct format is DP#.#####.00#"
        )


# %% functions to get the change / issue logs


def get_change_log_df(dpid, token=None):
    """
    Retrieves the change log for a NEON data product.

    Args:
        dpid (str): The NEON data product ID.
        token (str, optional): The NEON API token. Defaults to None.

    Returns:
        change_log_df: A DataFrame containing the changeLogs for the provided dpid.
        columns of the dataframe are: 'id', 'parentIssueID', 'issueDate',
        'resolvedDate', 'dateRangeStart', 'dateRangeEnd', 'locationAffected',
        'issue', 'resolution'
    """
    req = get_api(
        api_url=f"https://data.neonscience.org/api/v0/products/{dpid}", token=token
    )
    if req is None:
        logging.info(f"Error in metadata retrieval for {dpid}. Issue log not found.")
        return None
    all_product_info = pd.json_normalize(req.json()["data"])
    change_log_df = pd.DataFrame(all_product_info["changeLogs"][0])
    if not change_log_df.empty:
        change_log_df = change_log_df.astype({'id': 'int32', 'parentIssueID': 'str', 'issueDate': 'str',
                                              'resolvedDate': 'str', 'dateRangeStart': 'str', 'dateRangeEnd': 'str',
                                              'locationAffected': 'str', 'issue': 'str', 'resolution': 'str'})

    return change_log_df


def get_eddy_issue_log(dpid, token=None):
    """
    Retrieves the issue log for bundled eddy covariance data products.

    Args:
        dpid (str): The NEON data product ID.
        token (str, optional): The NEON API token. Defaults to None.

    Returns:
        eddy_issue_log_df: A DataFrame containing the bundled eddy covariance issue logs, including the data product id of the sub-data products.
        columns of the dataframe are: 'dpid', 'id', 'parentIssueID', 'issueDate', 'resolvedDate', 'dateRangeStart',
               'dateRangeEnd', 'locationAffected', 'issue', 'resolution'
    """

    bundle_dps = [
        "DP1.00007.001",
        "DP1.00010.001",
        "DP1.00034.001",
        "DP1.00035.001",
        "DP1.00036.001",
        "DP1.00037.001",
        "DP1.00099.001",
        "DP1.00100.001",
        "DP2.00008.001",
        "DP2.00009.001",
        "DP2.00024.001",
        "DP3.00008.001",
        "DP3.00009.001",
        "DP3.00010.001",
        "DP4.00002.001",
        "DP4.00007.001",
        "DP4.00067.001",
        "DP4.00137.001",
        "DP4.00201.001",
        "DP4.00200.001",
    ]

    eddy_issue_log_list = []

    for dp in bundle_dps:
        change_log_df = get_change_log_df(dp, token=token)
        if change_log_df is not None and not change_log_df.empty:
            change_log_df.insert(loc=0, column="dpid", value=dp)
            eddy_issue_log_list.append(change_log_df)

    eddy_issue_log_df = pd.concat(eddy_issue_log_list, ignore_index=True)
    eddy_issue_log_df = eddy_issue_log_df.drop_duplicates()

    return eddy_issue_log_df


def get_issue_log(dpid, token=None):
    """
    Retrieves the issue log for any NEON data products. Bundled eddy covariance data products have an additional column of the sub-data product id.

    Args:
        dpid: str
            The NEON data product ID.

        token: str
            User-specific API token from data.neonscience.org user account. See
            https://data.neonscience.org/data-api/rate-limiting/ for details about
            API rate limits and user tokens. If omitted, download uses the public rate limit.

    Returns:
        issue_log_df: A pandas DataFrame containing the changeLogs for the provided dpid.
        columns of the bundled eddy data frame are: 'dpid', 'id',
        'parentIssueID', 'issueDate', 'resolvedDate', 'dateRangeStart',
        'dateRangeEnd', 'locationAffected', 'issue', 'resolution';
        all other data products have the same columns minus 'dpid'

    Example
    -------
    Get the issue log for Breeding landbird point counts (DP1.10003.001)

    >>> birdiss = get_issue_log(dpid="DP1.10003.001")

    """
    # raise value error and print message if dpid isn't formatted as expected
    validate_dpid(dpid)

    if dpid == "DP4.00200.001":
        issue_log_df = get_eddy_issue_log(dpid, token)
    else:
        issue_log_df = get_change_log_df(dpid, token)

    return issue_log_df



# -*- coding: utf-8 -*-
"""
Created on 14 Aug 2024

@author: Claire Lunch

Unit tests for stack_by_table()

"""

# import required packages
from src.neonutilities.unzip_and_stack import stack_by_table
import pandas as pd


def test_stack_by_table_mini():
    """
    Test that stack_by_table() stacks the example product correctly. Example includes site-date, site-all, and lab-all.
    Test works but leaves unzipped items in the repo. Need to figure out how to remove
    """
    litterlst = stack_by_table("./testdata/NEON_litterfall.zip", savepath="envt", progress=False)
    litterchk = pd.read_csv("./testdata/NEON_litterfall_baseline/stackedFiles/ltr_litterLignin.csv")
    assert all(litterchk == litterlst["ltr_litterLignin"]) is True



from .api_helpers import get_api
from .api_helpers import get_api_headers
from .api_helpers import get_zip_urls
from .api_helpers import download_urls
from .api_helpers import download_file
from .api_helpers import token_date
from .api_helpers import token_check
from .metadata_helpers import get_recent



# -*- coding: utf-8 -*-
"""
Created on 14 Mar 2024

@author: Claire Lunch

Unit tests for zips_by_product()

Mocking is not used here, tests access API resources.

"""

# import required packages
from src.neonutilities.tabular_download import zips_by_product
import pytest
import logging


def test_zips_by_product_dpid():
    """
    Test that the zips_by_product() function errors correctly for an invalid DPID
    """
    with pytest.raises(ValueError) as exc_info:
        zips_by_product(
            dpid="DP1.444.001", site="NIWO", startdate="2012-01", enddate="2022-12"
        )
    assert (
        str(exc_info.value)
        == "DP1.444.001 is not a properly formatted NEON data product ID. The correct format is DP#.#####.00#"
    )


def test_zips_by_product_site(caplog):
    """
    Test that the zips_by_product() function returns correct message for no data at particular site
    """
    caplog.set_level(logging.INFO)

    zips_by_product(
        dpid="DP1.10003.001",
        site=["OKSR", "ARIK"],
        startdate="2012-01",
        enddate="2022-12",
    )

    assert any(
        "There are no NEON DP1.10003.001 data at the selected sites." in record.message
        for record in caplog.records
    )


def test_zips_by_product_cloud():
    """
    Test that running in cloud mode returns the correct list of files
    """
    murls = zips_by_product(
        dpid="DP1.10003.001",
        site=["NIWO", "PUUM"],
        startdate="2019-01",
        enddate="2019-12",
        release="RELEASE-2024",
        check_size=False,
        cloud_mode=True,
    )
    lst = [
        "https://storage.googleapis.com/neon-publication/NEON.DOM.SITE.DP1.10003.001/NIWO/20190701T000000--20190801T000000/basic/NEON.D13.NIWO.DP1.10003.001.brd_perpoint.2019-07.basic.20231227T192510Z.csv",
        "https://storage.googleapis.com/neon-publication/release/tag/RELEASE-2024/NEON.DOM.SITE.DP1.10003.001/NIWO/20190701T000000--20190801T000000/basic/NEON.D13.NIWO.DP1.10003.001.EML.20190703-20190713.20240127T000425Z.xml",
        "https://storage.googleapis.com/neon-publication/NEON.DOM.SITE.DP1.10003.001/NIWO/20190701T000000--20190801T000000/basic/NEON.D13.NIWO.DP1.10003.001.variables.20231227T192510Z.csv",
        "https://storage.googleapis.com/neon-publication/NEON.DOM.SITE.DP1.10003.001/NIWO/20190701T000000--20190801T000000/basic/NEON.D13.NIWO.DP0.10003.001.validation.20231227T192510Z.csv",
        "https://storage.googleapis.com/neon-publication/NEON.DOM.SITE.DP1.10003.001/NIWO/20190701T000000--20190801T000000/basic/NEON.D13.NIWO.DP1.10003.001.brd_countdata.2019-07.basic.20231227T192510Z.csv",
        "https://storage.googleapis.com/neon-publication/release/tag/RELEASE-2024/NEON.DOM.SITE.DP1.10003.001/NIWO/20190701T000000--20190801T000000/basic/NEON.D13.NIWO.DP1.10003.001.readme.20240127T000425Z.txt",
        "https://storage.googleapis.com/neon-publication/NEON.DOM.SITE.DP1.10003.001/NIWO/20190701T000000--20190801T000000/basic/NEON.D13.NIWO.DP0.10003.001.categoricalCodes.20231227T192510Z.csv",
    ]
    assert murls[0] == lst


def test_zips_by_product_avg():
    """
    Test that download by averaging interval returns the correct list of files
    """
    murls = zips_by_product(
        dpid="DP1.00005.001",
        site=["NIWO", "PUUM"],
        startdate="2022-06",
        enddate="2022-07",
        timeindex=30,
        check_size=False,
        progress=False,
        release="RELEASE-2024",
        cloud_mode=True,
    )
    assert len(murls) == 2
    ml = murls[0]
    assert len(ml) == 29
    assert (
        "https://storage.googleapis.com/neon-publication/release/tag/RELEASE-2024/NEON.DOM.SITE.DP1.00005.001/NIWO/20220701T000000--20220801T000000/basic/NEON.D13.NIWO.DP1.00005.001.readme.20240127T000425Z.txt"
        in ml
    )
    assert list(murls[1].values())[0] == "RELEASE-2024"



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import pandas as pd
import pyarrow as pa
from pyarrow import dataset
from pyarrow import fs
import zipfile
import platform
import glob
import re
import warnings
import importlib_resources
from tqdm import tqdm
from .tabular_download import zips_by_product
from .get_issue_log import get_issue_log
from .citation import get_citation
from .helper_mods.api_helpers import readme_url
from .helper_mods.api_helpers import get_api
from .read_table_neon import get_variables, cast_table_neon
from . import __resources__
import logging

logging.basicConfig(level=logging.INFO, format="%(message)s")

varschema = pa.schema(
    [
        ("table", pa.string()),
        ("fieldName", pa.string()),
        ("description", pa.string()),
        ("dataType", pa.string()),
        ("units", pa.string()),
        ("sampleCode", pa.string()),
        ("downloadPkg", pa.string()),
        ("pubFormat", pa.string()),
        ("primaryKey", pa.string()),
        ("categoricalCodeName", pa.string()),
    ]
)


def unzip_zipfile(zippath):
    """
    Unzip a zip file either at just the top level or recursively through the file.

    Parameters
    --------
    zippath: The filepath of the input file.

    Return
    --------
    A list of unzipped files to be used in stack_by_table.

    Example
    --------
    unzip_zipfile(zippath=filepath)

    Created on Tue Mar 5 2024

    @author: Zachary Nickerson

    Updated Feb 2025 to use "with" so that zip files properly close out
    and fix error handling for Windows filepath character length limits

    """

    # Error handling on inputs
    if zippath[-4:] in [".zip", ".ZIP"]:
        outpath = os.path.dirname(zippath)
        level = "all"
    else:
        outpath = zippath
        level = "in"

    if level == "all":
        with zipfile.ZipFile(zippath, "r") as zip_ref:
            tl = zip_ref.namelist()
            # print('tl:',tl)
            # Construct full paths as they will be after extraction
            # this is used in stack_by_table, eg. # litterlst = nu.stack_by_table("./testdata/NEON_litterfall.zip", savepath="envt")
            full_extracted_paths = [
                os.path.join(os.path.abspath(zippath).replace(".zip", ""), zipname)
                for zipname in tl
            ]
            # print('full_extracted_paths:',full_extracted_paths)
            # print('len(full_extracted_paths):',[len(x) for x in full_extracted_paths])

            # Error handling for filepath character length limitations on Windows
            # This is a warning
            if platform.system() == "Windows" and any(
                len(x) > 260 for x in full_extracted_paths
            ):
                longest_path = max(full_extracted_paths, key=len)
                warnings.warn(
                    "Filepaths on Windows are limited to 260 characters. "
                    "Attempting to extract a filepath that is > 260 characters long. "
                    "Move your working or savepath directory closer to the root directory or enable "
                    "long path support in Windows."
                )

            try:
                # Unzip file and get list of zips within (this will still fail if paths are too long on Windows)
                zip_ref.extractall(path=outpath)
                zps = glob.glob(zippath[:-4] + "/*.zip")
                # If there are more zips within parent zip file, unzip those as well
                # this might be deprecated (shouldn't happen any  more). level as an input might also be deprecated
                # leaving in in case anything changes in tabular data
                if len(zps) > 0:
                    print(
                        "Multiple zip files are contained within the parent zip file. Need an example to properly code this up.\n"
                    )
            except FileNotFoundError as e:
                print(e)
                raise OSError(
                    "Filepaths on Windows are limited to 260 characters. "
                    f"Attempting to extract a filepath that is {len(longest_path)} characters long. "
                    "Move your working or savepath directory closer to the root directory or enable "
                    "long path support in Windows."
                )

    if level == "in":
        zps = glob.glob(outpath + "/*.zip")

        for i in range(0, len(zps)):
            with zipfile.ZipFile(zps[i], "r") as zip_refi:
                tl = zip_refi.namelist()

                # Construct full paths as they will be after extraction
                full_extracted_paths = [
                    os.path.join(zps[i].replace(".zip", ""), zipname) for zipname in tl
                ]
                longest_path = max(full_extracted_paths, key=len)
                # print('full extracted paths:',full_extracted_paths)
                # print('len(full_extracted_paths):',[len(x) for x in full_extracted_paths])

                # Warning for filepath character length limitations on Windows
                if platform.system() == "Windows" and any(
                    len(x) > 260 for x in full_extracted_paths
                ):
                    warnings.warn(
                        "Filepaths on Windows are limited to 260 characters. "
                        "Attempting to extract a filepath that is > 260 characters long. "
                        "Move your working or savepath directory closer to the root directory "
                        "or enable long path support in Windows.",
                        UserWarning,
                    )

                try:
                    outpathi = zps[i][:-4]
                    zip_refi.extractall(path=outpathi)
                except FileNotFoundError as e:
                    raise OSError(
                        "ERROR: Filepaths on Windows are limited to 260 characters. "
                        f"Attempting to extract a filepath that is {len(longest_path)} characters long. "
                        "Move your working or savepath directory closer to the root directory or enable "
                        "long path support in Windows."
                    )
                    print(e)
            os.remove(zps[i])

    return None


def find_datatables(folder, f_names=True):
    """

    Find data tables

    Parameters
    --------
    folder: The filepath location of the unzipped NEON download package folder.
    f_names: Full names - if true, then return the full file names including enclosing folders - if false, return only the file names

    Return
    --------
    A data frame of file names

    Example
    --------
    filenames = find_datatables(folder=folder, f_names=False)
    see stack_data_files_parallel

    Created on Wed Apr 17 2024

    @author: Zachary Nickerson
    """

    # get all .csv files in the folder and its subfolders
    csv_files = glob.glob(os.path.join(folder, "**", "*.csv"), recursive=True)

    # if fnames is True, return the full file paths; otherwise, return the base names
    if f_names:
        return csv_files
    else:
        return [os.path.basename(f) for f in csv_files]


def get_recent_publication(filepaths):
    """

    Returns the most recent files for those that do not need stacking

    Parameters
    --------
    filepaths: List of file paths

    Return
    --------
    The filepath of the file with the most recent publication date

    Created on Wed Apr 17 2024

    @author: Zachary Nickerson
    """

    # extract the publication dates from the file paths
    pub_dates = [
        re.search(r"20\d{2}\d{2}\d{2}", os.path.basename(f)) for f in filepaths
    ]
    pub_dates = [m.group(0) for m in pub_dates if m is not None]

    # get the most recent publication date
    recent_pub_date = max(pub_dates)

    # get the file paths that include the most recent publication date
    recent_files = [f for f in filepaths if recent_pub_date in os.path.basename(f)]

    return recent_files


def string_schema(v):
    """

    Assign all variables to string. Fallback option if table read with true schema fails

    Parameters
    --------
    v: A pandas table containing variable definitions

    Return
    --------
    A pyarrow schema for all string data types based on the variable names

    Created on Oct 29 2024

    @author: Claire Lunch
    """

    for i in range(0, len(v)):
        nm = v.fieldName[i]
        if i == 0:
            vschema = pa.schema([(nm, pa.string())])
        else:
            nfield = pa.field(nm, pa.string())
            vschema = vschema.append(nfield)

    return vschema


def unknown_string_schema(v):
    """

    Assign all variables to string. Fallback option if variables file can't be found

    Parameters
    --------
    v: A list of column names

    Return
    --------
    A pyarrow schema for all string data types based on the column names

    Created on Oct 29 2024

    @author: Claire Lunch
    """

    for i in range(0, len(v)):
        nm = v[i]
        if i == 0:
            vschema = pa.schema([(nm, pa.string())])
        else:
            nfield = pa.field(nm, pa.string())
            vschema = vschema.append(nfield)

    return vschema


def table_type_formats(flname):
    """

    Small helper function to define mapping of table types to file name formats

    Parameters
    --------
    flname: A file name, split into components

    Return
    --------
    A table type

    Created on 6 May 2024

    @author: Claire Lunch
    """

    flen = len(flname)
    if flen <= 6:
        return "lab"
    else:
        ymr = re.compile("[0-9]{4}-[0-9]{2}")
        if any([f for f in flname if ymr.search(f)]):
            return "site-date"
        else:
            return "site-all"


def find_table_types(datatables):
    """

    Find unique data tables and their types

    Parameters
    --------
    datatables: A list of data files

    Return
    --------
    An array of unique table names and their types

    Created on 3 May 2024

    @author: Claire Lunch
    """

    dt = [os.path.basename(d) for d in datatables]
    splitnames = [d.split(".") for d in dt]
    td = []
    for i in range(0, len(splitnames)):
        for j in range(2, len(splitnames[i])):
            s = splitnames[i][j]
            if not s == "sensor_positions" and not s == "science_review_flags":
                if "_" not in s:
                    continue
                else:
                    ss = str.replace(s, "_pub", "")
                    td.append(ss)

    if len(td) == 0:
        logging.info(
            "No data tables found, only metadata. Try downloading expanded package, and check availability on the NEON data portal."
        )
        return
    else:
        tn = list(set(td))

    tt = {}
    for k in range(0, len(tn)):
        tnk = tn[k]
        tnkr = re.compile("^" + tnk + "$|^" + tnk + "_pub$")
        tnknames = [trnn for trnn in splitnames for tr in trnn if tnkr.search(tr)]
        ttklist = list(map(table_type_formats, tnknames))
        ttk = list(set(ttklist))
        if len(ttk) > 1:
            raise ValueError(
                f"In files to be stacked, table {tnk} has been published under conflicting schedules. To avoid this problem, either work only with released data, or stack released and provisional data separately."
            )
            return
        else:
            tt[tnk] = ttk[0]

    return tt


def find_lab_names(flpths):
    """

    Small helper function to find names of labs from filenames of lap-specific files

    Parameters
    --------
    flpths: A list of file paths

    Return
    --------
    A list of unique lab names

    Created on 5 June 2024

    @author: Claire Lunch
    """

    flnms = [os.path.basename(f) for f in flpths]
    splitnames = [f.split(".") for f in flnms]
    labnames = [f[1] for f in splitnames]
    return list(set(labnames))


def find_sites(flpths):
    """

    Small helper function to find sites from file names

    Parameters
    --------
    flpths: A list of file paths

    Return
    --------
    A list of unique sites

    Created on 5 June 2024

    @author: Claire Lunch
    """

    flnms = [os.path.basename(f) for f in flpths]
    sr = re.compile("[.][A-Z]{4}[.]")
    sites = [sr.search(f).group(0) for f in flnms]
    sites = list(set(sites))
    sites = [re.sub(pattern="[.]", repl="", string=s) for s in sites]
    return sites


def remove_srf_dups(srftab):
    """

    Small helper function to remove duplicates and updated records from science review flag tables

    Parameters
    --------
    srftab: A pandas table of SRF records

    Return
    --------
    A de-duplicated pandas table of SRF records

    Created on 6 June 2024

    @author: Claire Lunch
    """

    idset = list(set(srftab["srfID"]))
    srfsublist = []
    for i in idset:
        srfsubi = srftab[srftab.srfID == i]
        indi = srfsubi.idxmax()["lastUpdateDateTime"]
        srfsublist.append(srfsubi.loc[indi])
    srfsub = pd.DataFrame(srfsublist)
    return srfsub


def align_sp_cols(sptab):
    """

    Small helper function to align column names between old and new sensor positions tables

    Parameters
    --------
    sptab: A pandas table of sensor positions records

    Return
    --------
    A pandas table of sensor positions records with only the current column names

    Created on 16 Sept 2024

    @author: Claire Lunch
    """

    oldcols = {
        "name": "sensorLocationID",
        "description": "sensorLocationDescription",
        "start": "positionStartDateTime",
        "end": "positionEndDateTime",
        "referenceName": "referenceLocationID",
        "referenceDescription": "referenceLocationIDDescription",
        "referenceStart": "referenceLocationIDStartDateTime",
        "referenceEnd": "referenceLocationIDEndDateTime",
        "referenceLatitude": "locationReferenceLatitude",
        "referenceLongitude": "locationReferenceLongitude",
        "referenceElevation": "locationReferenceElevation",
    }
    for k in list(oldcols.keys()):
        if all(sptab[k].isna()):
            sptab.drop(columns=k, inplace=True)
        else:
            sptab[oldcols[k]] = sptab[oldcols[k]].fillna(sptab[k])
            sptab.drop(columns=k, inplace=True)

    return sptab


def sort_dat(pdata):
    """

    Small helper function to sort data rows by location and date

    Parameters
    --------
    pdata: A pandas table of data records

    Return
    --------
    A pandas table of data records sorted by location and date

    Created on 17 Oct 2024

    @author: Claire Lunch
    """

    # sort the table by site, then HOR/VER, then date, all ascending
    pcols = pdata.columns.to_list()
    datevar = None
    if "collectDate" in pcols:
        datevar = "collectDate"
    else:
        if "endDate" in pcols:
            datevar = "endDate"
        else:
            if "startDate" in pcols:
                datevar = "startDate"
            else:
                if "date" in pcols:
                    datevar = "date"
                else:
                    if "endDateTime" in pcols:
                        datevar = "endDateTime"
                    else:
                        if "startDateTime" in pcols:
                            datevar = "startDateTime"
                        else:
                            if "yearMonth" in pcols:
                                datevar = "yearMonth"
                            else:
                                if "year" in pcols:
                                    datevar = "year"
    if "horizontalPosition" not in pcols:
        try:
            if datevar is None:
                pdata.sort_values(
                    by=["siteID"], ascending=[True], inplace=True, ignore_index=True
                )
            else:
                pdata.sort_values(
                    by=["siteID", datevar],
                    ascending=[True, True],
                    inplace=True,
                    ignore_index=True,
                )
        except Exception:
            try:
                pdata.sort_values(
                    by=[datevar], ascending=[True], inplace=True, ignore_index=True
                )
            except Exception:
                pass

    else:
        try:
            if datevar is None:
                pdata.sort_values(
                    by=["siteID", "horizontalPosition", "verticalPosition"],
                    ascending=[True, True, True],
                    inplace=True,
                    ignore_index=True,
                )
            else:
                pdata.sort_values(
                    by=["siteID", "horizontalPosition", "verticalPosition", datevar],
                    ascending=[True, True, True, True],
                    inplace=True,
                    ignore_index=True,
                )
        except Exception:
            pass

    return pdata


def stack_frame_files(framefiles, dpid, seqtyp=None, cloud_mode=False):
    """

    Helper function to stack "data frame" files. These files do not go through the normal
    publication process, they are stored and published as a fixed unit. NEON uses this
    workflow for very large tabular files that can't be handled by the standard OS
    data pipeline.

    Parameters
    --------
    framefiles: A list of filepaths pointing to the data frame files
    dpid: The data product identifier of the product being stacked
    seqtyp: For microbe community data, sequence type 16S or ITS
    cloud_mode: Use cloud mode to transfer files cloud-to-cloud? Defaults to False.

    Return
    --------
    A stacked pandas table

    Created on 24 Oct 2024

    @author: Claire Lunch
    """

    # no variables files for these, use custom files in package resources
    frame_file_file = (
        importlib_resources.files(__resources__) / "frame_file_variables.csv"
    )
    frame_file_variables = pd.read_csv(frame_file_file, index_col=None)
    # v = pd.concat([v, frame_file_variables], ignore_index=True)

    fdict = {
        "DP1.30012.001": "FSP",
        "DP1.10081.001": "MCC",
        "DP1.20086.001": "MCC",
        "DP1.20141.001": "MCC",
        "DP1.20190.001": "REA",
        "DP1.20193.001": "REA",
        "DP1.10081.002": "MCT",
        "DP1.20086.002": "MCT",
        "DP1.20141.002": "MCT",
        "DP4.00132.001": "BAT",
    }

    fvars = pa.Table.from_pandas(frame_file_variables)
    ftab = fvars.filter(pa.compute.field("table") == fdict[dpid])

    fpkgvar = ftab.to_pandas()
    fschema = get_variables(fpkgvar)

    if cloud_mode:
        gcs = fs.GcsFileSystem(anonymous=True)
        framebuckets = [
            re.sub(pattern="https://storage.neonscience.org/", repl="", string=b)
            for b in framefiles
        ]
        fdat = dataset.dataset(
            source=framebuckets, filesystem=gcs, format="csv", schema=fschema
        )
    else:
        fdat = dataset.dataset(source=framefiles, format="csv", schema=fschema)

    try:
        fdattab = fdat.to_table()
    except Exception:
        stringschema = unknown_string_schema(fdat.head(num_rows=0).column_names)
        if cloud_mode:
            fdat = dataset.dataset(
                source=framebuckets, filesystem=gcs, format="csv", schema=stringschema
            )
        else:
            fdat = dataset.dataset(source=framefiles, format="csv", schema=stringschema)
        fdattab = fdat.to_table()
        logging.info(
            "Large file schema did not match expectations; all variable types set to string."
        )

    fpdat = fdattab.to_pandas()

    nm = "per_sample"

    if dpid == "DP1.20190.001":
        nm = "rea_conductivityRawData"
    elif dpid == "DP1.20193.001":
        nm = "sbd_conductivityRawData"
    elif dpid == "DP1.30012.001":
        nm = "fsp_rawSpectra"
    elif dpid == "DP4.00132.001":
        nm = "bat_processedSonarFile"
    elif dpid == "DP1.10081.001":
        nm = f"mcc_soilPerSampleTaxonomy_{seqtyp}"
    elif dpid == "DP1.20086.001":
        nm = f"mcc_benthicPerSampleTaxonomy_{seqtyp}"
    elif dpid == "DP1.20141.001":
        nm = f"mcc_surfaceWaterPerSampleTaxonomy_{seqtyp}"
    elif dpid == "DP1.10081.002":
        nm = f"mct_soilPerSampleTaxonomy_{seqtyp}"
    elif dpid == "DP1.20086.002":
        nm = f"mct_benthicPerSampleTaxonomy_{seqtyp}"
    elif dpid == "DP1.20141.002":
        nm = f"mct_surfaceWaterPerSampleTaxonomy_{seqtyp}"

    return {"frmdat": fpdat, "frmnm": nm}


def format_readme(readmetab, tables):
    """

    Remove site-specific information from the most recently published readme file in the download, and add generic information about the neonutilities download.

    Parameters
    --------
    readmetab:  A readme document as a pandas table
    tables: A list of data tables included in download

    Return
    --------
    A modified readme file formatted to remove site-specific information and include standard information about the neonutilities download

    Created on Fri Jul 12 2024

    @author: Zachary Nickerson
    """

    rd = readmetab
    if len(tables) > 0:
        # replace query specific text
        dpackind = rd[0].str.contains("CONTENTS").idxmax()
        rd.loc[
            dpackind + 2, 0
        ] = f"This data product contains up to {len(tables)} data tables:"
        rd.loc[dpackind + 3 : dpackind + 3 + len(tables) - 1, 0] = tables
        rd.loc[
            dpackind + 3 + len(tables), 0
        ] = "If data are unavailable for the particular sites and dates queried, some tables may be absent."
        qind = rd[0].str.contains("QUERY").idxmax()
        downpackind = rd[0].str.contains("Basic download package").idxmax()

        # Remove specific rows
        remove_indices = (
            list(range(qind, dpackind))
            + list(range(dpackind + 4 + len(tables), downpackind))
            + rd.index[rd[0].str.contains("Date-Time")].tolist()
        )
        remove_indices = [index for index in remove_indices if index in list(rd.index)]
        rd = rd.drop(remove_indices)

        # add disclaimer
        disclaimer = pd.DataFrame(
            {
                0: [
                    "###################################",
                    "########### Disclaimer ############",
                    "This is the most recent readme publication based on all site-date combinations used during stackByTable.",
                    "Information specific to the query, including sites and dates, has been removed. The remaining content reflects general metadata for the data product.",
                    "##################################",
                ]
            }
        )
        rd = pd.concat([disclaimer, rd], ignore_index=True)
    return rd


def stack_data_files_parallel(folder, 
                              package, 
                              dpid, 
                              progress=True, 
                              cloud_mode=False,
                              datasetq=False):
    """

    Join data files in a unzipped NEON data package by table type

    Parameters
    --------
    folder: The filepath location of the unzipped NEON download package folder.
    package: basic or expanded data package
    dpid: Data product ID of product to stack.
    progress: Should a progress bar be displayed?
    cloud_mode: Use cloud mode to transfer files cloud-to-cloud? If used, stack_by_table() expects a list of file urls as input. Defaults to False.
    datasetq: Return an arrow dataset for a single table? Defaults to False.

    Return
    --------
    One file for each table type is created and written.

    Created on Tue Apr 2 2024

    @author: Zachary Nickerson
    """

    releases = []

    dpnum = dpid[4:9]

    if cloud_mode:
        filenames = [os.path.basename(f) for f in folder[0]]
        filepaths = folder[0]
        gcs = fs.GcsFileSystem(anonymous=True)
    else:
        # Get filenames without full path
        filenames = find_datatables(folder=folder, f_names=False)

        # Get filenames with full path
        filepaths = find_datatables(folder=folder, f_names=True)

    # dictionary for outputs
    stacklist = {}

    # handle per-sample (data frame) tables separately
    if (
        dpid
        in [
            "DP1.30012.001",
            "DP1.10081.001",
            "DP1.20086.001",
            "DP1.20141.001",
            "DP1.20190.001",
            "DP1.20193.001",
            "DP1.10081.002",
            "DP1.20086.002",
            "DP1.20141.002",
            "DP4.00132.001",
        ]
        and len([f for f in filenames if not f.startswith("NEON.")]) > 0
    ):
        framefiles = [
            f for f in filepaths if not os.path.basename(f).startswith("NEON.")
        ]
        filepaths = [f for f in filepaths if os.path.basename(f).startswith("NEON.")]
        filenames = [f for f in filenames if os.path.basename(f).startswith("NEON.")]

        # stack frame files
        if progress:
            logging.info(
               f"Stacking NEON {dpid} per-sample files. These files may be very large; download data in smaller subsets if performance problems are encountered.\n"
            )

        # subset microbe community data by taxonomic group
        # and stack both sets
        if dpid in [
            "DP1.10081.001",
            "DP1.20086.001",
            "DP1.20141.001",
            "DP1.10081.002",
            "DP1.20086.002",
            "DP1.20141.002",
        ]:
            bacteriafiles = [b for b in framefiles if re.search("[_]16S[_]", b)]
            fungifiles = [b for b in framefiles if re.search("[_]ITS[_]", b)]

            if len(bacteriafiles) > 0:
                fpdat16 = stack_frame_files(
                    bacteriafiles, dpid=dpid, seqtyp="16S", cloud_mode=cloud_mode
                )
                stacklist[fpdat16["frmnm"]] = fpdat16["frmdat"]

            if len(fungifiles) > 0:
                fpdatIT = stack_frame_files(
                    fungifiles, dpid=dpid, seqtyp="ITS", cloud_mode=cloud_mode
                )
                stacklist[fpdatIT["frmnm"]] = fpdatIT["frmdat"]

        else:
            fpdat = stack_frame_files(
                framefiles, dpid=dpid, seqtyp=None, cloud_mode=cloud_mode
            )
            stacklist[fpdat["frmnm"]] = fpdat["frmdat"]

    # make a dictionary, where filenames are the keys to the filepath values
    filelist = dict(zip(filenames, filepaths))

    datafls = filelist

    # if there are no datafiles, exit
    if len(datafls) == 0:
        print(f"No NEON {dpid} data files are present in specified file path.\n")
        return None

    # if there is one or more than one file, stack files

    # get table types
    table_types = find_table_types(filenames)
    if any(re.search("sensor_positions", path) for path in filepaths):
        table_types["sensor_positions"] = "site-all"
    if any(re.search("science_review_flags", path) for path in filepaths):
        table_types["science_review_flags"] = "site-date"
    tables = list(table_types.keys())

    # metadata files
    # get variables and validation files using the most recent publication date
    if any(re.search("variables.20", path) for path in filepaths):
        varpath = get_recent_publication(
            [path for path in filepaths if "variables.20" in path]
        )[0]
        if cloud_mode:
            vp = dataset.dataset(
                source=re.sub("https://storage.googleapis.com/", "", varpath),
                filesystem=gcs,
                format="csv",
                schema=varschema,
            )
            va = vp.to_table()
            v = va.to_pandas()
        else:
            v = pd.read_csv(varpath, sep=",")

        # if science review flags are present but missing from variables file, add variables
        if "science_review_flags" not in list(v["table"]):
            if any("science_review_flags" in path for path in filepaths):
                science_review_file = (
                    importlib_resources.files(__resources__)
                    / "science_review_variables.csv"
                )
                science_review_variables = pd.read_csv(
                    science_review_file, index_col=None
                )
                v = pd.concat([v, science_review_variables], ignore_index=True)

        # if sensor positions are present but missing from variables file, add variables
        if any("sensor_positions" in path for path in filepaths):
            sensor_positions_map = (
                importlib_resources.files(__resources__)
                / "sensor_positions_variables_mapping.csv"
            )
            sensor_positions_internal_variables = pd.read_csv(
                sensor_positions_map, index_col=None
            )
            if "sensor_positions" not in list(v["table"]):
                sensor_positions_file = (
                    importlib_resources.files(__resources__)
                    / "sensor_positions_variables.csv"
                )
                sensor_positions_variables = pd.read_csv(
                    sensor_positions_file, index_col=None
                )
                v = pd.concat([v, sensor_positions_variables], ignore_index=True)

        # save the variables file
        vlist = {k: v for k, v in v.groupby("table")}
        stacklist[f"variables_{dpnum}"] = v

    # get validation file
    if any(re.search("validation", path) for path in filepaths):
        valpath = get_recent_publication(
            [path for path in filepaths if "validation" in path]
        )[0]
        if cloud_mode:
            vp = dataset.dataset(
                source=re.sub("https://storage.googleapis.com/", "", valpath),
                filesystem=gcs,
                format="csv",
                schema=None,
            )
            va = vp.to_table()
            val = va.to_pandas()
        else:
            val = pd.read_csv(valpath, sep=",")
        stacklist[f"validation_{dpnum}"] = val

    # get categoricalCodes file
    if any(re.search("categoricalCodes", path) for path in filepaths):
        ccpath = get_recent_publication(
            [path for path in filepaths if "categoricalCodes" in path]
        )[0]
        if cloud_mode:
            cp = dataset.dataset(
                source=re.sub("https://storage.googleapis.com/", "", ccpath),
                filesystem=gcs,
                format="csv",
                schema=None,
            )
            ca = cp.to_table()
            cc = ca.to_pandas()
        else:
            cc = pd.read_csv(ccpath, sep=",")
        stacklist[f"categoricalCodes_{dpnum}"] = cc

    # get readme file
    if cloud_mode:
        readmefiles = filepaths
    else:
        readmefiles = glob.glob(os.path.join(folder, "**", "*.txt"), recursive=True)
    if any(re.search("readme.20", path) for path in readmefiles):
        readmepath = get_recent_publication(
            [path for path in readmefiles if "readme.20" in path]
        )[0]
        rd = None
        if cloud_mode:
            try:
                rd = readme_url(readmepath)
            except Exception:
                pass
        else:
            try:
                rd = pd.read_table(readmepath, delimiter="\t", header=None)
            except Exception:
                pass
        if rd is None:
            pass
        else:
            rd = format_readme(rd, tables)
            # save the readme
            stacklist[f"readme_{dpnum}"] = rd

    # stack tables according to types
    if progress:
        logging.info("Stacking data files")
    arrowvars = pa.Table.from_pandas(stacklist[f"variables_{dpnum}"])
    for j in tqdm(tables, disable=not progress):
        # create schema from variables file, for only this table and package
        vtab = arrowvars.filter(pa.compute.field("table") == j)
        if len(vtab) == 0:
            vtab = arrowvars.filter(pa.compute.field("table") == j + "_pub")
        if j == "sensor_positions":
            vtab = pa.Table.from_pandas(sensor_positions_internal_variables)

        if package == "basic":
            vtabpkg = vtab.filter(pa.compute.field("downloadPkg") == "basic")
        else:
            vtabpkg = vtab

        tablepkgvar = vtabpkg.to_pandas()
        if len(tablepkgvar) == 0:
            # set to string if variables file can't be found
            tableschema = None
            logging.info(
                f"NEON {dpid} variables file not found for table {j}. Data types will be inferred if possible."
            )
        else:
            tableschema = get_variables(tablepkgvar)

        # subset the list of files to the relevant table
        tabler = re.compile("[.]" + j + "[.]|[.]" + j + "_pub[.]")
        tablepaths = [f for f in filepaths if tabler.search(f)]

        # subset the list of files for lab-specific tables:
        # get the most recent file from each lab
        if table_types[j] == "lab":
            labs = find_lab_names(tablepaths)
            labrecent = list()
            for k in labs:
                labr = re.compile(k)
                labpaths = [f for f in tablepaths if labr.search(f)]
                labrecent.append(get_recent_publication(labpaths)[0])
            tablepaths = labrecent

        # subset the list of files for site-all tables:
        # get the most recent file from each site
        if table_types[j] == "site-all":
            sites = find_sites(tablepaths)
            siterecent = list()
            for k in sites:
                sr = re.compile(k)
                sitepaths = [f for f in tablepaths if sr.search(f)]
                siterecent.append(get_recent_publication(sitepaths)[0])
            tablepaths = siterecent

        # read data and append file names
        if cloud_mode:
            tablebuckets = [
                re.sub(pattern="https://storage.googleapis.com/", repl="", string=b)
                for b in tablepaths
            ]
            dat = dataset.dataset(
                source=tablebuckets, filesystem=gcs, format="csv", schema=tableschema
            )
        else:
            dat = dataset.dataset(source=tablepaths, format="csv", schema=tableschema)

        if tableschema is None:
            cols = dat.head(num_rows=0).column_names
        else:
            cols = tableschema.names
        cols.append("__filename")

        # attempt to stack to table. if it fails, stack as all string fields and warn
        stringset = False
        try:
            dattab = dat.to_table(columns=cols)
        except Exception:
            try:
                if tableschema is None:
                    stringschema = unknown_string_schema(
                        dat.head(num_rows=0).column_names
                    )
                else:
                    stringschema = string_schema(tablepkgvar)
                if cloud_mode:
                    dat = dataset.dataset(
                        source=tablebuckets,
                        filesystem=gcs,
                        format="csv",
                        schema=stringschema,
                    )
                else:
                    dat = dataset.dataset(
                        source=tablepaths, format="csv", schema=stringschema
                    )
                dattab = dat.to_table(columns=cols)
                logging.info(
                    f"Table {j} schema did not match data; all variable types set to string. Data type casting will be attempted after stacking step."
                )
                stringset = True
            except Exception:
                logging.info(
                    f"Failed to stack table {j}. Check input data and variables file."
                )
                continue
            
        # for dataset query, done here
        if datasetq:
            return dat
        else:
            pdat = dattab.to_pandas()

        if stringset:
            try:
                pdat = cast_table_neon(pdat, tablepkgvar)
            except Exception:
                logging.info(
                    f"Data type casting failed for table {j}. Variable types set to string."
                )

        # append publication date
        pubr = re.compile("20[0-9]{6}T[0-9]{6}Z")
        pubval = [pubr.search(os.path.basename(p)).group(0) for p in pdat["__filename"]]
        pdat = pdat.assign(publicationDate=pubval)

        # append release tag
        if cloud_mode:
            pdat["release"] = pdat["__filename"].map(folder[1])
            releases.append(list(set(folder[1].values())))
        else:
            pubrelr = re.compile("20[0-9]{6}T[0-9]{6}Z\\..*\\/")
            pubrelval = [pubrelr.search(p).group(0) for p in pdat["__filename"]]
            relval = [re.sub(".*\\.", "", s) for s in pubrelval]
            relval = [re.sub("\\/", "", s) for s in relval]
            pdat = pdat.assign(release=relval)
            releases.append(list(set(relval)))

        # append fields to variables file
        if f"variables_{dpnum}" in stacklist.keys():
            added_fields_file = (
                importlib_resources.files(__resources__) / "added_fields.csv"
            )
            added_fields = pd.read_csv(added_fields_file, index_col=None)
            added_fields_all = added_fields[-2:]
            added_fields_all.insert(0, "table", j)
            try:
                vlist[j] = pd.concat([vlist[j], added_fields_all], ignore_index=True)
            except Exception:
                pass

        # for IS products, append domainID, siteID, HOR, VER
        if "siteID" not in pdat.columns.to_list() and not table_types[j] == "lab":
            dr = re.compile("D[0-2]{1}[0-9]{1}")
            domval = [dr.search(d).group(0) for d in pdat["__filename"]]
            pdat.insert(0, "domainID", domval)

            sr = re.compile("D[0-9]{2}[.][A-Z]{4}[.]")
            sitel = [sr.search(s).group(0) for s in pdat["__filename"]]
            siteval = [
                re.sub(pattern="D[0-9]{2}[.]|[.]", repl="", string=s) for s in sitel
            ]
            pdat.insert(1, "siteID", siteval)

            if j != "sensor_positions":
                locr = re.compile("[.][0-9]{3}[.][0-9]{3}[.][0-9]{3}[.][0-9]{3}[.]|[.][0-9]{3}[.][0-9]{3}[.][0-9]{3}[.][0-9]{2}[A-Z]{1}[.]")
                indtemp = [locr.search(l) for l in pdat["__filename"]]
                if None in indtemp:
                    pdat = sort_dat(pdat)
                else:
                    indxs = [lt.group(0) for lt in indtemp]
                    hor = [indx[5:8] for indx in indxs]
                    ver = [indx[9:12] for indx in indxs]
                    pdat.insert(2, "horizontalPosition", hor)
                    pdat.insert(3, "verticalPosition", ver)

                    # sort table rows
                    pdat = sort_dat(pdat)

                # append fields to variables file
                if f"variables_{dpnum}" in stacklist.keys():
                    added_fields_IS = added_fields[0:4]
                    added_fields_IS.insert(0, "table", j)
                    try:
                        vlist[j] = pd.concat(
                            [added_fields_IS, vlist[j]], ignore_index=True
                        )
                    except Exception:
                        pass

        else:
            # for OS tables, sort by site and date
            pdat = sort_dat(pdat)

        # for SRF files, remove duplicates and modified records
        if j == "science_review_flags":
            pdat = remove_srf_dups(pdat)

        # for sensor position files, align column names
        if j == "sensor_positions":
            pdat = align_sp_cols(pdat)

        # remove filename column
        pdat = pdat.drop(columns=["__filename"])
        
        # add table to list
        if j == "science_review_flags" or j == "sensor_positions":
            stacklist[f"{j}_{dpnum}"] = pdat
        else:
            stacklist[j] = pdat

    # final variables file
    stacklist[f"variables_{dpnum}"] = pd.concat(vlist, ignore_index=True)

    # get issue log table
    # token omitted here since it's not otherwise used in stacking functions
    # consider a runLocal option, like in R stackEddy()
    try:
        stacklist[f"issueLog_{dpnum}"] = get_issue_log(dpid=dpid, token=None)
    except Exception:
        pass

    # get relevant citation(s)
    try:
        releases = sum(releases, [])
        releases = list(set(releases))
        if "PROVISIONAL" in releases:
            try:
                stacklist[f"citation_{dpnum}_PROVISIONAL"] = get_citation(
                    dpid=dpid, release="PROVISIONAL"
                )
            except Exception:
                pass
        relr = re.compile("RELEASE-20[0-9]{2}")
        rs = [relr.search(r).group(0) for r in releases if relr.search(r)]
        if len(rs) == 1:
            stacklist[f"citation_{dpnum}_{rs[0]}"] = get_citation(
                dpid=dpid, release=rs[0]
            )
        if len(rs) > 1:
            logging.info(
                f"Multiple NEON data releases were stacked together for {dpid}. This is not appropriate, check your input data."
            )
    except Exception:
        pass

    return stacklist


def stack_by_table(
    filepath, savepath=None, 
    save_unzipped_files=False, 
    progress=True, 
    cloud_mode=False,
    datasetq=False,
):
    """

    Join data files in a zipped or unzipped NEON data package by table type.
    This function can be used on a zip file downloaded from the NEON data portal or
    on a set of files downloaded by zips_by_product().

    Parameters
    ------------------
    filepath: str
        The location of the zip file or downloaded files.

    savepath: str, optional
        The location to save the output files to. If omitted, output files will be
        saved in the same location as the input file.

    save_unzipped_files: bool, optional
        Should the unzipped monthly data folders be retained? Defaults to False.

    progress: bool, optional
        Should the function display progress bars as it runs? Defaults to True.

    cloud_mode: bool, optional
        Use cloud mode to transfer files cloud-to-cloud? If used, stack_by_table()
        expects a list of file urls as input. Defaults to False; in general this
        option should be used via load_by_product(), in which stack_by_table() is a
        helper function.
        
    datasetq: bool, optional
        Should the function return an arrow dataset for a single table, instead 
        of a set of stacked tables? Defaults to False.


    Return
    -------------------
    All files are unzipped and one file for each table type is created and written.
    If savepath="envt" is specified, output is a named list of tables; otherwise,
    function output is null and files are saved to the location specified.

    Example
    ------------------
    To stack PAR data (DP1.00024.001) downloaded from the NEON data portal

    >>> pardat = stack_by_table("/filepath/NEON_par.zip")

    Notes
    --------

    Windows Path Length Limitations:
    When using this function, you may encounter path length limitations  on Windows systems.
    Windows has a default maximum path length of 260 characters, which can cause download and unzip
    functions to fail if this limit is exceeded. If the file path exceeds 260 characters on a Windows system,
    the package will issue a warning.You can choose to ignore or filter these warnings using Python's warnings
    module if you prefer not to see them. If the function is unable to unzip a folder due to path length
    limitations, an OSError will be raised.

    Created on Tue Mar 5 2024

    @author: Zachary Nickerson
    """

    # determine contents of filepath and unzip as needed
    if cloud_mode:
        files = filepath[0]
    else:
        # Is the filepath input a zip file or an unzipped file?
        if filepath[-4:] in [".zip", ".ZIP"]:
            folder = False
        else:
            folder = True

        # Get list of files nested (and/or zipped) in filepath
        if not folder:
            zip_ref = zipfile.ZipFile(filepath, "r")
            files = zip_ref.namelist()
        else:
            files = glob.glob(filepath + "/**", recursive=True)

    # Error handling if there are no standardized NEON Portal data tables in the list of files
    if not any(re.search(r"NEON.D[0-9]{2}.[A-Z]{4}.", x) for x in files):
        logging.info("NEON data files are not present in the specified filepath.")
        return

    # Determine dpid
    # this regexpr allows for REV = .001 or .002
    dpid_listlist = []
    for f in range(len(files)):
        dpid_listlist.append(
            re.findall(re.compile("DP[1-4][.][0-9]{5}[.]00[1-2]{1}"), files[f])
        )
    dpid = [x for dpid_list in dpid_listlist for x in dpid_list]
    dpid = list(set(dpid))
    if not len(dpid) == 1:
        logging.info(
            "NEON Data product ID could not be determined. Check that filepath contains data files, from a single NEON data product."
        )
        return
    else:
        dpid = dpid[0]

    # Determine download package
    package_listlist = []
    for f in range(len(files)):
        package_listlist.append(re.findall(re.compile("basic|expanded"), files[f]))
    package = [x for package_list in package_listlist for x in package_list]
    package = list(set(package))
    if "expanded" in package:
        package = "expanded"
    else:
        package = "basic"

    # Error message for AOP data
    if dpid[4] == "3" and not dpid == "DP1.30012.001":
        logging.info(
            "This is a NEON AOP data product, files cannot be stacked. Use by_file_aop() or by_tile_aop() to download data."
        )
        return

    # Error messafe for SAE data
    if dpid == "DP4.00200.001":
        logging.info(
            "This NEON eddy covariance data product is in HDF5 format. Stack using the stackEddy() function in the R package version of neonUtilities."
        )
        return

    # Exceptions for digital hemispheric photos
    if dpid == "DP1.10017.001" and package == "expanded":
        save_unzipped_files = True
        logging.info(
            "Note: NEON Digital hemispheric photos (in NEF format) cannot be stacked; only the CSV metadata files will be stacked."
        )

    # Warning about all sensor soil data
    # Test and modify the file length for the alert, this should be a lot better with arrow
    if dpid in ["DP1.00094.001", "DP1.00041.001"] and len(files) > 24:
        logging.info(
            "Warning! Attempting to stack NEON soil sensor data. Note that due to the number of soil sensors at each site, data volume is very high for these data. Consider dividing data processing into chunks and/or using a high-performance system."
        )

    # If all checks pass, unzip and stack files
    if cloud_mode:
        stackedlist = stack_data_files_parallel(
            folder=filepath,
            package=package,
            dpid=dpid,
            progress=progress,
            cloud_mode=True,
            datasetq=datasetq,
        )

    else:
        # If the filepath is a zip file
        if not folder:
            unzip_zipfile(zippath=filepath)
            stackpath = filepath[:-4]

        # If the filepath is a directory
        if folder:
            if any(".zip" in file for file in files):
                unzip_zipfile(zippath=filepath)
            stackpath = filepath

        # Stack the files
        stackedlist = stack_data_files_parallel(
            folder=stackpath, package=package, dpid=dpid, 
            progress=progress, cloud_mode=False, datasetq=False,
        )

        # delete input files
        if not save_unzipped_files:
            ufl = glob.glob(stackpath + "/**.*/*", recursive=True)
            for fl in ufl:
                try:
                    os.remove(fl)
                except Exception:
                    pass
            dirlist = glob.glob(stackpath + "/*", recursive=True)
            for d in dirlist:
                try:
                    os.rmdir(d)
                except Exception:
                    pass
            if os.listdir(stackpath) == [".DS_Store"]:
                os.remove(stackpath + "/.DS_Store")
            if os.listdir(stackpath) == []:
                os.rmdir(stackpath)

    # Check if stackedlist is None (eg. if there is a failure in downloading the data due to long paths on windows)
    if stackedlist is None:
        logging.info("ERROR! Stacking failed due to previous errors.")
        return None
    
    # if returning an arrow dataset, done here
    if datasetq:
        return stackedlist

    # sort the dictionary of tables
    mk = list(stackedlist.keys())
    mk.sort()
    stackedlistsort = {i: stackedlist[i] for i in mk}

    # write files to path if requested
    if savepath == "envt":
        return stackedlistsort
    else:
        if savepath is None:
            stacked_files_dir = os.path.join(stackpath, "stackedFiles")

        elif savepath != "envt":
            stacked_files_dir = os.path.join(savepath, "stackedFiles")

        if not os.path.exists(stacked_files_dir):
            os.makedirs(stacked_files_dir)

        for k in stackedlistsort.keys():
            tk = stackedlistsort[k]
            if "citation" in k:
                with open(
                    f"{stacked_files_dir}/{k}.txt", mode="w+", encoding="utf-8"
                ) as f:
                    f.write(tk)
            else:
                if "readme" in k:
                    tk.to_csv(f"{stacked_files_dir}/{k}.txt", sep="\t", index=False)
                else:
                    # Write numeric data to up to 15 digits, rounding to
                    # the precision just below the maximum. This mimics the
                    # default behavior in R, to ensure the two languages'
                    # versions of neonUtilities match. For an interesting
                    # discussion of the issues around this, see
                    # https://github.com/pandas-dev/pandas/issues/16452
                    tk.to_csv(
                        f"{stacked_files_dir}/{k}.csv",
                        index=False,
                        float_format="%.15g",
                    )

        return None


def load_by_product(
    dpid,
    site="all",
    startdate=None,
    enddate=None,
    package="basic",
    release="current",
    timeindex="all",
    tabl="all",
    check_size=True,
    include_provisional=False,
    cloud_mode=False,
    progress=True,
    token=None,
):
    """
    This function downloads product-site-month data package files from NEON, unzips
    and stacks the data files, and loads to the environment.

    Parameters
    ----------------
    dpid: str
        Data product identifier in the form DP#.#####.###

    site: str
        Either the string 'all', or one or more 4-letter NEON site codes. Defaults to 'all'.

    startdate: str, optional
        Earliest date of data to download, in the form YYYY-MM

    enddate: str, optional
        Latest date of data to download, in the form YYYY-MM

    package: str, optional
        Download package to access, either basic or expanded. Defaults to 'basic'.

    release: str, optional
        Data release to download. Defaults to the most recent release.

    timeindex: str, optional
        Either 'all', or the time index of data to download, in minutes. Only applicable to sensor (IS) data. Defaults to 'all'.

    tabl: str, optional
        Either 'all', or the name of a single data table to download. Only applicable to observational (OS) data. Defaults to 'all'.

    check_size: bool, optional
        True or False, should the user approve the total file size before downloading? Defaults to True. When working in batch mode, or other non-interactive workflow, use check_size=False.

    include_provisional: bool, optional
        Should Provisional data be returned in the download? Defaults to False. See
        https://www.neonscience.org/data-samples/data-management/data-revisions-releases
        for details on the difference between provisional and released data.

    cloud_mode: bool, optional
        Use cloud mode to transfer files cloud-to-cloud? Should only be used if the destination location is in the cloud. Defaults to False.

    progress: bool, optional
        Should the function display progress bars as it runs? Defaults to True.

    token: str, optional
        User-specific API token from data.neonscience.org user account. See
        https://data.neonscience.org/data-api/rate-limiting/ for details about
        API rate limits and user tokens. If omitted, download uses the public rate limit.

    Return
    ---------------
    A dictionary of data and metadata tables, as pandas tables.

    Example
    ---------------
    Download water quality data from COMO (Como Creek) in 2018

    >>> wq = load_by_product(dpid="DP1.20288.001", site="COMO",
                             startdate="2018-01", enddate="2018-12",
                             token=None)

    Notes
    --------
    Windows Path Length Limitations:
    When using this function, you may encounter path length limitations  on Windows systems.
    Windows has a default maximum path length of 260 characters, which can cause download and unzip
    functions to fail if this limit is exceeded. If the file path exceeds 260 characters on a Windows system,
    the package will issue a warning.You can choose to ignore or filter these warnings using Python's warnings
    module if you prefer not to see them. If the function is unable to unzip a folder due to path length
    limitations, an OSError will be raised.

    Created on June 12 2024

    @author: Claire Lunch
    """

    savepath = os.getcwd()

    if cloud_mode:
        flist = zips_by_product(
            dpid=dpid,
            site=site,
            startdate=startdate,
            enddate=enddate,
            package=package,
            release=release,
            timeindex=timeindex,
            tabl=tabl,
            check_size=check_size,
            include_provisional=include_provisional,
            cloud_mode=True,
            progress=progress,
            token=token,
            savepath=savepath,
        )

        outlist = stack_by_table(
            filepath=flist,
            savepath="envt",
            cloud_mode=True,
            save_unzipped_files=False,
            progress=progress,
        )

    else:
        zips_by_product(
            dpid=dpid,
            site=site,
            startdate=startdate,
            enddate=enddate,
            package=package,
            release=release,
            timeindex=timeindex,
            tabl=tabl,
            check_size=check_size,
            include_provisional=include_provisional,
            progress=progress,
            token=token,
            savepath=savepath,
        )

        stackpath = savepath + "/filesToStack" + dpid[4:9] + "/"

        outlist = stack_by_table(
            filepath=stackpath,
            savepath="envt",
            save_unzipped_files=False,
            progress=progress,
        )

    return outlist


def dataset_query(
    dpid,
    site="all",
    startdate=None,
    enddate=None,
    package="basic",
    release="current",
    tabl=None,
    hor=None,
    ver=None,
    include_provisional=False,
    token=None,
):
    """
    This function uses the query endpoint of the NEON API to find the full 
    list of files for a given data product, release, site(s), and date range, 
    then turns them into an arrow dataset.

    Parameters
    ----------------
    dpid: str
        Data product identifier in the form DP#.#####.###

    site: str
        Either the string 'all', or one or more 4-letter NEON site codes. Defaults to 'all'.

    startdate: str, optional
        Earliest date of data to include in dataset, in the form YYYY-MM

    enddate: str, optional
        Latest date of data to include in dataset, in the form YYYY-MM

    package: str, optional
        Download package to access, either basic or expanded. Defaults to 'basic'.

    release: str, optional
        Data release to access. Defaults to the most recent release.

    tabl: str
        The name of a single data table to access.

    hor: str, optional
        The horizontal index of data to access. Only applicable to sensor (IS) data.

    ver: str, optional
        The vertical index of data to access. Only applicable to sensor (IS) data.

    include_provisional: bool, optional
        Should Provisional data be returned in the download? Defaults to False. See
        https://www.neonscience.org/data-samples/data-management/data-revisions-releases
        for details on the difference between provisional and released data.

    token: str, optional
        User-specific API token from data.neonscience.org user account. See
        https://data.neonscience.org/data-api/rate-limiting/ for details about
        API rate limits and user tokens. If omitted, download uses the public rate limit.

    Return
    ---------------
    An arrow dataset for the data requested.

    Example
    ---------------
    Create a dataset of mammal trap data from HARV (Harvard Forest) in 2018

    >>> mamds = dataset_query(dpid="DP1.10072.001", site="HARV",
                              tabl="mam_pertrapnight", package="expanded",
                              startdate="2018-01", enddate="2018-12",
                              token=None)

    Created on August 12 2025

    @author: Claire Lunch
    """

    # check that needed inputs are present and data product is tabular
    # consider turning opening checks in zips_by_product() into a function
    
    # query the /products endpoint for the product requested
    if release == "current" or release == "PROVISIONAL":
        prodreq = get_api(
            api_url="https://data.neonscience.org/api/v0/products/" + dpid, token=token
        )
    else:
        prodreq = get_api(
            api_url="https://data.neonscience.org/api/v0/products/"
            + dpid
            + "?release="
            + release,
            token=token,
        )
        
        if prodreq is None:
            if release == "LATEST":
                logging.info(
                    f"No data found for NEON data product {dpid}. LATEST data requested; check that token is valid for LATEST access."
                )
                return
            else:
                logging.info(
                    f"No data found for NEON data product {dpid}."
                )
                return

    avail = prodreq.json()
    pubtype = avail["data"]["productPublicationFormatType"]

    if pubtype=="AOP Data Product Type":
        raise ValueError(
            f"{dpid} is a remote sensing data product. Remote sensing data can't be queried using this function."
            )
        
    if tabl is None:
        raise ValueError(
            "Table name (tabl=) is a required input to this function."
            )
        
    if pubtype in ["TOS Data Product Type","TOS-structured TIS Data Product Type"]:
        raise ValueError(
            f"{dpid} is an observational data product. hor and ver are not valid inputs for this data product type."
            )
        
    if pubtype in ["TIS Data Product Type","AIS Data Product Type"]:
        if dpid in ["DP4.00200.001",
                   "DP1.00007.001","DP1.00010.001","DP1.00034.001","DP1.00035.001",
                   "DP1.00036.001","DP1.00037.001","DP1.00099.001","DP1.00100.001",
                   "DP2.00008.001","DP2.00009.001","DP2.00024.001","DP3.00008.001",
                   "DP3.00009.001","DP3.00010.001","DP4.00002.001","DP4.00007.001",
                   "DP4.00067.001","DP4.00137.001","DP4.00201.001","DP1.00030.001"]:
            raise ValueError(
                f"{dpid} is an eddy covariance data product and can't be queried using this function."
                )
        else:
            if bool(re.search(pattern="^ais[_]", string=tabl)):
                tabl = tabl
            else:
                if site=="all" or not isinstance(site, str):
                    raise ValueError(
                        f"{dpid} is a sensor data product and can only be queried at a single site using this function. If you need data at multiple sites, run multiple queries or use loadByProduct()."
                        )
                if hor is None or ver is None:
                    raise ValueError(
                        f"{dpid} is a sensor data product. hor and ver indices must be provided to use this query function."
                        )
                if len(hor)!=3 or len(ver)!=3 or not isinstance(hor, str) or not isinstance(ver, str):
                    raise ValueError(
                        "hor and ver must be 3-digit codes in string format, and must each be a single code. Only one sensor location can be queried."
                        )
    
    # get list of files
    flist = zips_by_product(
        dpid=dpid,
        site=site,
        startdate=startdate,
        enddate=enddate,
        package=package,
        release=release,
        tabl=tabl,
        check_size=False,
        include_provisional=include_provisional,
        cloud_mode=True,
        token=token,
        savepath=None,
    )
    
    # if IS data, subset by hor and ver
    # keep the variables file for schema creation
    if ver is not None:
        hvr = re.compile("[.]00[0-9]{1}[.]" + hor + "[.]" + ver + "[.][0-9]{2}[A-Z0-9]{1}[.]|variables")
        d0 = [r for r in flist[0] if hvr.search(r)]
        d1 = {r: flist[1][r] for r in flist[1] if hvr.search(r)}
    
        flist[0] = d0
        flist[1] = d1
    
    # send to stacking function to turn into dataset
    # stacking stops right before the dataset-to-table step
    outds = stack_by_table(
        filepath=flist,
        savepath="envt",
        cloud_mode=True,
        save_unzipped_files=False,
        progress=False,
        datasetq=True,
    )

    return outds


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import re
import os


def get_recent(fl_set, fltype):
    """

    Given a set of file names and urls for NEON data files, returns the url for the most recently published item of type=fltype, based on publication datetime stamp in the file name.

    Parameters
    --------
    fl_set: A set of file names and urls for NEON data files, in the format returned via the data endpoint of the NEON API.
    fltype: The type of file to be returned (e.g. variables, sensor positions).

    Return
    --------
    The most recently published object (contains file name, url, size).

    Created on Apr 18 2024

    @author: Claire Lunch
    """

    # subset to files of specified type
    ty = re.compile(fltype)
    flt = [f for f in fl_set if ty.search(f["name"])]

    if len(flt) == 0:
        return None

    # get max date stamp in subset
    pdater = re.compile("[0-9]{8}T[0-9]{6}Z")
    flvar = [pdater.search(os.path.basename(f["name"])).group(0) for f in flt]

    if len(flvar) == 0:
        return None

    recdate = max(flvar)

    # get url matching max date stamp
    maxr = re.compile(recdate)
    flmax = [f for f in flt if maxr.search(os.path.basename(f["name"]))]

    return [flmax[0]]


def convert_byte_size(size_bytes):
    """
    This function converts the file size in bytes to a more readable format.
    It converts bytes to Kilobytes (KB), Megabytes (MB), Gigabytes (GB), or Terabytes (TB)
    depending on the size of the input. This uses 1024 bytes as the base for conversion.

    Parameters
    --------
    size_bytes: int or float
        The file size in bytes. It should be a non-negative number.

    Returns
    --------
    str
        A string that represents the file size in a more readable format.
        The format includes the size number followed by the size unit (KB, MB, GB, or TB).

    Raises
    --------
    None

    Examples
    --------
    >>> convert_byte_size(5000)
    '4.9 KB'

    >>> convert_byte_size(2000000)
    '1.9 MB'

    >>> convert_byte_size(3000000000)
    '2.8 GB'

    >>> convert_byte_size(4000000000000)
    '3.6 TB'"""
    si_prefix = [[40, "T"], [30, "G"], [20, "M"], [10, "K"], [0, ""]]

    for si_row in si_prefix:
        if size_bytes >= 2 ** si_row[0]:
            break

    return f"{(size_bytes / 2 ** si_row[0]):.1f} {si_row[1]}B"



# -*- coding: utf-8 -*-
"""
Created on Wed Feb 21 17:00:19 2024

@author: bhass

Installation dependencies:
    conda install pytest
    conda install parameterized

Unit tests for by_file_aop and by_tile_aop.

These mainly test the functions' output messages for invalid inputs,
collocated sites, data not found, and provisional availability scenarios

More complete integration tests are included in the NEONScience/nu_python_testing repository, which is currently private.

Notes:
- These tests require an internet connection to run, as they are actually making API calls
- Paramaterization with the @parameterized.expand decorator allows for testing different sets of inputs with the same test
    This requires the parameterized package (pip install parameterized)
- The @patch('builtins.input', return_value='n') decorator mocks the user input to say "n" (no) to avoid downloading data


"""

# import required packages
from datetime import datetime
import os
import pytest
import unittest
from unittest.mock import patch
from parameterized import parameterized

# import neon utilities functions that are being tested
from src.neonutilities.aop_download import (
    by_file_aop,
    by_tile_aop,
)

# read in token from os.environ (requires the token to be set)
token = os.environ.get("NEON_TOKEN")

# Test invalid inputs and make sure informational messages display correctly
# This contains similar tests for by_file_aop and by_tile_aop,
# with some additional tests for extra functionality in by_tile_aop


class TestByFileAOP(unittest.TestCase):
    def setUp(self):
        """Set up before each test"""
        # these are valid inputs, invalid inputs for testing logs / messages are supplied within each test
        self.dpid = "DP3.30015.001"
        self.site = "MCRA"
        self.year = 2021

    def test_invalid_dpid_format(self):
        """
        Test that invalid dpid format raises ValueError and message displays correctly.
        """
        invalid_dpid = "DP1.30001"
        with self.assertRaises(
            ValueError,
            msg=f"{invalid_dpid} is not a properly formatted NEON data product ID. The correct format is DP#.#####.00#",
        ):
            by_file_aop(dpid=invalid_dpid, site=self.site, year=self.year)

#     def test_invalid_aop_dpid_pattern(self):
#         """
#         Test that invalid AOP dpid pattern raises a ValueError and message displays correctly.
#         """
#         invalid_aop_dpid = "DP1.20001.001"
#         with self.assertRaises(
#             ValueError,
#             msg=f"{invalid_aop_dpid} is not a valid NEON AOP data product ID. AOP products follow the format DP#.300##.00#",
#         ):
#             by_file_aop(dpid=invalid_aop_dpid, site=self.site, year=self.year)

#     def test_invalid_aop_dpid_suspended(self):
#         """
#         Test that entering a suspended AOP data product raises a ValueError and message displays correctly.
#         """
#         suspended_aop_dpid = "DP2.30016.001"
#         # ' Valid AOP DPIDs are '):
#         with self.assertRaises(
#             ValueError,
#             msg=f"NEON {suspended_aop_dpid} has been suspended and is not currently available, see https://www.neonscience.org/data-products/{suspended_aop_dpid} for more details.",
#         ):
#             by_file_aop(dpid=suspended_aop_dpid, site=self.site, year=self.year)

#     def test_check_field_spectra_dpid(self):
#         """
#         Test that providing field spectra dpid raises ValueError and message displays correctly.
#         """
#         field_spectra_dpid = "DP1.30012.001"
#         with self.assertRaises(
#             ValueError,
#             msg=f"NEON {field_spectra_dpid} is the Field spectral data product, which is published as tabular data. Use zipsByProduct() or loadByProduct() to download these data.",
#         ):
#             by_file_aop(dpid=field_spectra_dpid, site=self.site, year=self.year)

#     def test_invalid_site_format(self):
#         """
#         Test that invalid site format raises ValueError and message displays correctly.
#         """
#         invalid_site = "McRae"
#         with self.assertRaises(
#             ValueError,
#             msg=f"{invalid_site} is an invalid NEON site format. A four-letter NEON site code is required. NEON sites codes can be found here: https://www.neonscience.org/field-sites/explore-field-sites",
#         ):
#             by_file_aop(dpid=self.dpid, site=invalid_site, year=self.year)

#     def test_invalid_neon_site(self):
#         """
#         Test that an invalid NEON site code raises a ValueError and the message displays correctly.
#         """
#         invalid_site = "HOPD"
#         with self.assertRaises(
#             ValueError,
#             msg=f"{invalid_site} is not a valid NEON site code. A complete list of NEON site codes can be found here: https://www.neonscience.org/field-sites/explore-field-sites",
#         ):
#             by_file_aop(dpid=self.dpid, site=invalid_site, year=self.year)

#     def test_valid_year(self):
#         # Should not raise for a valid year
#         current_year = datetime.now().year
#         try:
#             by_file_aop(dpid=self.dpid, site=self.site, year=current_year)

#         except ValueError:
#             self.fail(
#                 "validate_year() raised ValueError unexpectedly for a valid year."
#             )

#     def test_invalid_year_out_of_range(self):
#         current_year = datetime.now().year
#         # Should raise for a year before 2012
#         with pytest.raises(
#             ValueError,
#             match=f"2011 is an invalid year. Year must be between 2012 and {current_year}.",
#         ):
#             by_file_aop(dpid=self.dpid, site=self.site, year=2011)
#         # Should raise for a year after the current yearAdd commentMore actions
#         future_year = datetime.now().year + 1
#         with pytest.raises(
#             ValueError,
#             match=f"{future_year} is an invalid year. Year must be between 2012 and {current_year}.",
#         ):
#             by_file_aop(dpid=self.dpid, site=self.site, year=future_year)

#     @parameterized.expand(
#         [
#             ("21",),
#             ("2021-05",),
#         ]
#     )
#     def test_invalid_year_format(self, year):
#         """
#         Test that invalid year format raises ValueError and message displays correctly.
#         """
#         with pytest.raises(
#             ValueError,
#             match=f"{year} is an invalid year. Year is required in the format '2017' or 2017, e.g. NEON AOP data are available from 2012 to present.",
#         ):
#             by_file_aop(dpid=self.dpid, site=self.site, year=year)

#     @parameterized.expand(
#         [
#             ("2022", "CHEQ", "STEI"),
#             ("2022", "TREE", "STEI"),
#             ("2021", "DCFS", "WOOD"),
#             ("2020", "KONA", "KONZ"),
#         ],
#         name_func=lambda f, n, p: f"{f.__name__}_{p.args[1]}_{p.args[0]}",
#     )
#     # the name_func displays a more descriptive test name when running the tests
#     @patch("builtins.input", return_value="n")
#     def test_collocated_terrestrial_site_message(
#         self, year, site, flightSite, input_mock
#     ):
#         """
#         Test application of the terrestrial collocated site lookup, and expected message display.
#         """
#         with self.assertLogs(level="INFO") as cm:
#             by_file_aop(dpid="DP3.30015.001", site=site, year=year, token=token)
#             self.assertIn(
#                 f"INFO:root:{site} is part of the NEON flight box for {flightSite}. Downloading data from {flightSite}.",
#                 cm.output,
#             )

#     @parameterized.expand(
#         [
#             ("2018", "BARC", "OSBS"),
#             ("2020", "COMO", "NIWO"),
#             # ("2021", "BLDE", "YELL"),
#             # ("2020", "KING", "KONZ"),
#         ],
#         name_func=lambda f, n, p: f"{f.__name__}_{p.args[1]}_{p.args[0]}",
#     )
#     @patch("builtins.input", return_value="n")
#     def test_collocated_aquatic_site_message(self, year, site, flightSite, input_mock):
#         """
#         Test application of the aquatic collocated site lookup, and expected message display.
#         """
#         with self.assertLogs(level="INFO") as cm:
#             by_file_aop(dpid=self.dpid, site=site, year=year, token=token)
#             self.assertIn(
#                 f"INFO:root:{site} is a NEON aquatic site and is sometimes included in the flight box for {flightSite}. Aquatic sites are not always included in the flight coverage every year.\nDownloading data from {flightSite}. Check data to confirm coverage of {site}.",
#                 cm.output,
#             )

#     def test_no_data_available_message(self):
#         """
#         Test that the by_file_aop() function returns the expected error log when no data is available for a selected site and year.
#         """
#         with self.assertLogs(level="INFO") as cm:
#             by_file_aop(dpid="DP3.30015.001", site=self.site, year=2020)
#             self.assertIn(
#                 f"INFO:root:There are no NEON {self.dpid} data available at the site {self.site} in 2020.\nTo display available dates for a given data product and site, use the function list_available_dates().",
#                 cm.output,
#             )

#     @patch("builtins.input", return_value="n")
#     def test_check_download_size_message(self, input_mock):
#         """
#         Test that download check_size message displays correctly.
#         """
#         result = by_file_aop(dpid=self.dpid, site=self.site, year=self.year)
#         # Check that the function asked for confirmation to download and prints expected message.
#         input_mock.assert_called_once_with(
#             "Continuing will download 128 NEON data files totaling approximately 93.1 MB. Do you want to proceed? (y/n) "
#         )
#         # Check that the function halted the download
#         self.assertEqual(result, None)

#     # Provisional scenarios

#     # provisional not included, and no data available
#     def test_all_provisional_no_data_available_message(self):
#         """
#         Test that the by_file_aop() function returns the expected message when include_provisional is set to False (default) but no data are available.
#         This has already run through the check that any data is available(eg. there is data at that site for the year provided)
#         """
#         with self.assertLogs(level="INFO") as cm:
#             by_file_aop(dpid="DP3.30015.001", site="WLOU", year=2024)
#             self.assertIn(
#                 "INFO:root:No NEON data files found. Available data may all be provisional. To download provisional data, use input parameter include_provisional=True.",
#                 cm.output,
#             )

#     # provisional included, and no data available
#     @patch("builtins.input", return_value="n")
#     def test_provisional_included_and_data_available_message(self, input_mock):
#         """
#         Test that the by_file_aop() function returns the expected message when include_provisional is set to False (default) but no data are available.
#         This has already run through the check that any data is available(eg. there is data at that site for the year provided)
#         """
#         with self.assertLogs(level="INFO") as cm:
#             by_file_aop(
#                 dpid="DP3.30015.001", site="WLOU", year=2024, include_provisional=True
#             )
#             self.assertIn(
#                 "INFO:root:Provisional NEON data are included. To exclude provisional data, use input parameter include_provisional=False.",
#                 cm.output,
#             )

    # overwrite warning when skip_if_exists is False
    # @patch("builtins.input", return_value="n")
    # def test_by_file_aop_overwrite_warning_when_skip_if_exists_false(self, input_mock):
    #     with self.assertLogs(level="INFO") as cm:
    #         by_file_aop(
    #             dpid="DP3.30015.001",
    #             site="MCRA",
    #             year="2022",
    #             skip_if_exists=False,
    #             overwrite="no",
    #         )
    #     self.assertIn(
    #         "WARNING: overwrite option only applies if skip_if_exists=True. By default, any existing files will be overwritten unless you select skip_if_exists=True and overwrite='no' or 'prompt' (default).",
    #         "\n".join(cm.output),
    #     )

    # other scenarios- check messages but don't download the data
    # provisional not included, and data available
    # provisional included, and data available


# class TestByTileAop(unittest.TestCase):
#     def setUp(self):
#         """Set up before each test"""
#         # these are valid inputs, invalid inputs for each tests are supplied within each test
#         self.dpid = "DP3.30015.001"
#         self.easting = 566000
#         self.northing = 4900000
#         self.site = "MCRA"
#         self.year = 2021

#     def test_invalid_dpid_format(self):
#         invalid_dpid = "DP1.30001"
#         with self.assertRaises(
#             ValueError,
#             msg=f"{invalid_dpid} is not a properly formatted NEON data product ID. The correct format is DP#.#####.00#",
#         ):
#             by_tile_aop(
#                 dpid=invalid_dpid,
#                 site=self.site,
#                 year=self.year,
#                 easting=self.easting,
#                 northing=self.northing,
#             )

#     def test_invalid_aop_l3_dpid(self):
#         """
#         Test that invalid AOP dpid raises a ValueError and message displays correctly.
#         """
#         invalid_aop_dpid = "DP1.30001.001"
#         with self.assertRaises(
#             ValueError,
#             msg=f"NEON {invalid_aop_dpid} is not a valid Level 3 AOP data product ID. Level 3 AOP products follow the format DP3.300##.00#",
#         ):
#             by_tile_aop(
#                 dpid=invalid_aop_dpid,
#                 site=self.site,
#                 year=self.year,
#                 easting=self.easting,
#                 northing=self.northing,
#             )

#     def test_check_field_spectra_dpid(self):
#         field_spectra_dpid = "DP1.30012.001"
#         with self.assertRaises(
#             ValueError,
#             msg=f"NEON {field_spectra_dpid} is the Field spectral data product, which is published as tabular data. Use zipsByProduct() or loadByProduct() to download these data.",
#         ):
#             by_tile_aop(
#                 dpid=field_spectra_dpid,
#                 site=self.site,
#                 year=self.year,
#                 easting=self.easting,
#                 northing=self.northing,
#             )

#     def test_invalid_site_format(self):
#         invalid_site = "McRae"
#         with self.assertRaises(
#             ValueError,
#             msg=f"{invalid_site} is an invalid NEON site format. A four-letter NEON site code is required. NEON sites codes can be found here: https://www.neonscience.org/field-sites/explore-field-sites",
#         ):
#             by_tile_aop(
#                 dpid=self.dpid,
#                 site=invalid_site,
#                 year=self.year,
#                 easting=self.easting,
#                 northing=self.northing,
#             )

#     def test_invalid_neon_site(self):
#         invalid_site = "ABBA"
#         with self.assertRaises(
#             ValueError,
#             msg=f"{invalid_site} is not a valid NEON site code. A complete list of NEON site codes can be found here: https://www.neonscience.org/field-sites/explore-field-sites",
#         ):
#             by_tile_aop(
#                 dpid=self.dpid,
#                 site=invalid_site,
#                 year=self.year,
#                 easting=self.easting,
#                 northing=self.northing,
#             )

#     @parameterized.expand(
#         [
#             ("21",),
#             ("2021-05",),
#         ]
#     )
#     def test_invalid_year_format(self, year):
#         """
#         Test that invalid year format raises ValueError and message displays correctly.
#         """
#         with pytest.raises(
#             ValueError,
#             match=f"{year} is an invalid year. Year is required in the format '2017' or 2017, e.g. NEON AOP data are available from 2012 to present.",
#         ):
#             by_tile_aop(
#                 dpid=self.dpid,
#                 site=self.site,
#                 year=year,
#                 easting=self.easting,
#                 northing=self.northing,
#             )

#     @parameterized.expand(
#         [
#             ("2022", "CHEQ", "STEI"),
#             ("2022", "TREE", "STEI"),
#             ("2021", "DCFS", "WOOD"),
#             ("2020", "KONA", "KONZ"),
#         ],
#         name_func=lambda f, n, p: f"{f.__name__}_{p.args[1]}_{p.args[0]}",
#     )
#     # the name_func displays a more descriptive test name when running the tests
#     @patch("builtins.input", return_value="n")
#     def test_collocated_terrestrial_site_message(
#         self, year, site, flightSite, input_mock
#     ):
#         """
#         Test application of the terrestrial collocated site lookup, and expected message display.
#         """
#         with self.assertLogs(level="INFO") as cm:
#             by_tile_aop(dpid=self.dpid, site=site, year=year, easting=[], northing=[])
#             self.assertIn(
#                 f"INFO:root:{site} is part of the NEON flight box for {flightSite}. Downloading data from {flightSite}.",
#                 cm.output,
#             )

#     @parameterized.expand(
#         [
#             ("2018", "BARC", "OSBS"),
#             ("2020", "COMO", "NIWO"),
#             # ("2021", "BLDE", "YELL"),
#             # ("2020", "KING", "KONZ"),
#         ],
#         name_func=lambda f, n, p: f"{f.__name__}_{p.args[1]}_{p.args[0]}",
#     )
#     @patch("builtins.input", return_value="n")
#     def test_collocated_aquatic_site_message(self, year, site, flightSite, input_mock):
#         """
#         Test application of the aquatic collocated site lookup, and expected message display.
#         """
#         with self.assertLogs(level="INFO") as cm:
#             by_tile_aop(dpid=self.dpid, site=site, year=year, easting=[], northing=[])
#             self.assertIn(
#                 f"INFO:root:{site} is a NEON aquatic site and is sometimes included in the flight box for {flightSite}. Aquatic sites are not always included in the flight coverage every year.\nDownloading data from {flightSite}. Check data to confirm coverage of {site}.",
#                 cm.output,
#             )

#     def test_no_data_available_message(self):
#         """
#         Test that the by_tile_aop() function returns the expected error log when no data is available for a selected site and year.
#         """
#         with self.assertLogs(level="INFO") as cm:
#             by_tile_aop(
#                 dpid="DP3.30015.001",
#                 site=self.site,
#                 year=2020,
#                 easting=self.easting,
#                 northing=self.northing,
#             )
#             self.assertIn(
#                 f"INFO:root:There are no NEON DP3.30015.001 data available at the site {self.site} in 2020.\nTo display available dates for a given data product and site, use the function list_available_dates().",
#                 cm.output,
#             )

#     def test_no_data_files_found_message(self):
#         """
#         Test that the by_tile_aop() function returns the expected error log when no data files are found.
#         """
#         with self.assertLogs(level="INFO") as cm:
#             by_tile_aop(
#                 dpid="DP3.30015.001",
#                 site=self.site,
#                 year=2020,
#                 easting=564000,
#                 northing=4900000,
#             )
#             self.assertIn(
#                 f"INFO:root:There are no NEON DP3.30015.001 data available at the site {self.site} in 2020.\nTo display available dates for a given data product and site, use the function list_available_dates().",
#                 cm.output,
#             )

#     @patch("builtins.input", return_value="n")
#     def test_check_download_size_message(self, input_mock):
#         """
#         Test that download check_size message displays correctly.
#         """
#         result = by_tile_aop(
#             dpid=self.dpid,
#             site=self.site,
#             year=self.year,
#             easting=self.easting,
#             northing=self.northing,
#         )
#         # Check that the function asked for confirmation to download and prints expected message.
#         input_mock.assert_called_once_with(
#             "Continuing will download 7 NEON data files totaling approximately 3.9 MB. Do you want to proceed? (y/n) "
#         )
#         # Check that the function halted the download
#         self.assertEqual(result, None)

#     # Provisional scenarios

#     # provisional not included, and no data available
#     # these tests should be mocked, as provisional data changes with each NEON Data Release
#     def test_all_provisional_no_data_available_message(self):
#         """
#         Test that the by_tile_aop() function returns the expected message when include_provisional is set to False (default) but no data are available.
#         This has already run through the check that any data is available(eg. there is data at that site for the year provided)
#         """
#         with self.assertLogs(level="INFO") as cm:
#             by_tile_aop(
#                 dpid=self.dpid,
#                 site="WLOU",
#                 year=2024,
#                 easting=self.easting,
#                 northing=self.northing,
#             )
#             self.assertIn(
#                 "INFO:root:No NEON data files found. Available data may all be provisional. To download provisional data, use input parameter include_provisional=True.",
#                 cm.output,
#             )

#     # provisional included, and no data available
#     @patch("builtins.input", return_value="n")
#     def test_provisional_included_and_data_available_message(self, input_mock):
#         """
#         Test that the by_file_aop() function returns the expected message when include_provisional is set to False (default) but no data are available.
#         This has already run through the check that any data is available(eg. there is data at that site for the year provided)
#         """
#         with self.assertLogs(level="INFO") as cm:
#             by_tile_aop(
#                 dpid=self.dpid,
#                 site=self.site,
#                 year=2023,
#                 include_provisional=True,
#                 easting=self.easting,
#                 northing=self.northing,
#             )
#             self.assertIn(
#                 "INFO:root:Provisional NEON data are included. To exclude provisional data, use input parameter include_provisional=False.",
#                 cm.output,
#             )

    # pyproj is part of the requirements, so this is not needed - this was carried over from the R package
    # @ patch('builtins.input', return_value='n')
    # @ patch('importlib.import_module')
    # @ patch('logging.info')
    # def test_pyproj_not_installed(self, logging_mock, import_module_mock, input_mock):
    #     # Setup the mock to raise ImportError when pyproj is attempted to be imported
    #     import_module_mock.side_effect = ImportError(
    #         "No module named 'pyproj'")

    #     by_tile_aop(dpid="DP3.30015.001", site='BLAN', year=2022,
    #                 easting=243758.81, northing=4330667.37,
    #                 include_provisional=False, verbose=True)

    #     # Check if logging.info was called with the correct message
    #     # There will also be a message about the include_provisional, so can't use assert_called_once_with
    #     logging_mock.assert_any_call(
    #         "Package pyproj is required for this function to work at the BLAN site. Install and re-try."
    #     )

    @patch("builtins.input", return_value="n")
    def test_blan_utm_info_message(self, input_mock):
        """
        Test that the by_tile_aop() function returns the expected message about UTM zone conversion when BLAN is the site.
        """
        with self.assertLogs(level="INFO") as cm:
            by_tile_aop(
                dpid="DP3.30015.001",
                site="BLAN",
                year=2022,
                easting=243758.81,
                northing=4330667.37,
                verbose=True,
            )
            self.assertIn(
                "INFO:root:Blandy (BLAN) plots include two UTM zones, flight data are all in 17N. "
                "Coordinates in UTM zone 18N have been converted to 17N to download the correct tiles. "
                "You will need to make the same conversion to connect airborne to ground data.",
                cm.output,
            )
            self.assertIn("INFO:root:UTM 17N Easting(s): 762717.81", cm.output)
            self.assertIn("INFO:root:UTM 17N Northing(s): 4330881.38", cm.output)
            self.assertIn(
                "INFO:root:UTM (x, y) lower-left coordinates of tiles to be downloaded:",
                cm.output,
            )
            self.assertIn("INFO:root:(762000, 4330000)", cm.output)